# Arithmetic Rules

import types
import typing as typ
import warnings

import numpy as np

import pyxu.abc.operator as pxo
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.runtime as pxrt
import pyxu.util as pxu


class Rule:
    """
    General arithmetic rule.

    This class defines default arithmetic rules applicable unless re-defined by sub-classes.
    """

    def op(self) -> pxt.OpT:
        """
        Returns
        -------
        op: OpT
            Synthesize operator.
        """
        raise NotImplementedError

    # Helper Methods ----------------------------------------------------------

    @staticmethod
    def _propagate_constants(op: pxt.OpT):
        # Propagate (diff-)Lipschitz constants forward via special call to
        # Rule()-overridden `estimate_[diff_]lipschitz()` methods.

        # Important: we write to _[diff_]lipschitz to not overwrite estimate_[diff_]lipschitz() methods.
        if op.has(pxo.Property.CAN_EVAL):
            op._lipschitz = op.estimate_lipschitz(__rule=True)
        if op.has(pxo.Property.DIFFERENTIABLE):
            op._diff_lipschitz = op.estimate_diff_lipschitz(__rule=True)

    # Default Arithmetic Methods ----------------------------------------------
    # Fallback on these when no simple form in terms of Rule.__init__() args is known.
    # If a method from Property.arithmetic_methods() is not listed here, then all Rule subclasses
    # provide an overloaded implementation.

    def __call__(self, arr: pxt.NDArray) -> pxt.NDArray:
        return self.apply(arr)

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = self.__class__.svdvals(self, **kwargs)
        return D

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = self.__class__.pinv(self, arr=arr, damp=damp, **kwargs)
        return out

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = self.__class__.trace(self, **kwargs)
        return tr


class ScaleRule(Rule):
    r"""
    Arithmetic rules for element-wise scaling: :math:`B(x) = \alpha A(x)`.

    Special Cases::

        \alpha = 0  => NullOp/NullFunc
        \alpha = 1  => self

    Else::

        |--------------------------|-------------|--------------------------------------------------------------------|
        |         Property         |  Preserved? |                     Arithmetic Update Rule(s)                      |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | CAN_EVAL                 | yes         | op_new.apply(arr) = op_old.apply(arr) * \alpha                     |
        |                          |             | op_new.lipschitz = op_old.lipschitz * abs(\alpha)                  |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | FUNCTIONAL               | yes         | op_new.asloss(\beta) = op_old.asloss(\beta) * \alpha               |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | PROXIMABLE               | \alpha > 0  | op_new.prox(arr, tau) = op_old.prox(arr, tau * \alpha)             |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | DIFFERENTIABLE           | yes         | op_new.jacobian(arr) = op_old.jacobian(arr) * \alpha               |
        |                          |             | op_new.diff_lipschitz = op_old.diff_lipschitz * abs(\alpha)        |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | DIFFERENTIABLE_FUNCTION  | yes         | op_new.grad(arr) = op_old.grad(arr) * \alpha                       |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | QUADRATIC                | \alpha > 0  | Q, c, t = op_old._quad_spec()                                      |
        |                          |             | op_new._quad_spec() = (\alpha * Q, \alpha * c, \alpha * t)         |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR                   | yes         | op_new.adjoint(arr) = op_old.adjoint(arr) * \alpha                 |
        |                          |             | op_new.asarray() = op_old.asarray() * \alpha                       |
        |                          |             | op_new.svdvals() = op_old.svdvals() * abs(\alpha)                  |
        |                          |             | op_new.pinv(x, damp) = op_old.pinv(x, damp / (\alpha**2)) / \alpha |
        |                          |             | op_new.gram() = op_old.gram() * (\alpha**2)                        |
        |                          |             | op_new.cogram() = op_old.cogram() * (\alpha**2)                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_SQUARE            | yes         | op_new.trace() = op_old.trace() * \alpha                           |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_NORMAL            | yes         |                                                                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_UNITARY           | \alpha = -1 |                                                                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_SELF_ADJOINT      | yes         |                                                                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_POSITIVE_DEFINITE | \alpha > 0  |                                                                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
        | LINEAR_IDEMPOTENT        | no          |                                                                    |
        |--------------------------|-------------|--------------------------------------------------------------------|
    """

    def __init__(self, op: pxt.OpT, cst: pxt.Real):
        super().__init__()
        self._op = op.squeeze()
        self._cst = float(cst)

    def op(self) -> pxt.OpT:
        if np.isclose(self._cst, 0):
            from pyxu.operator import NullOp

            op = NullOp(shape=self._op.shape).squeeze()
        elif np.isclose(self._cst, 1):
            op = self._op
        else:
            klass = self._infer_op_klass()
            op = klass(shape=self._op.shape)
            op._op = self._op  # embed for introspection
            op._cst = self._cst  # embed for introspection
            for p in op.properties():
                for name in p.arithmetic_methods():
                    func = getattr(self.__class__, name)
                    setattr(op, name, types.MethodType(func, op))
            self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        return ("scale", self._op, self._cst)

    def _infer_op_klass(self) -> pxt.OpC:
        preserved = {
            pxo.Property.CAN_EVAL,
            pxo.Property.FUNCTIONAL,
            pxo.Property.DIFFERENTIABLE,
            pxo.Property.DIFFERENTIABLE_FUNCTION,
            pxo.Property.LINEAR,
            pxo.Property.LINEAR_SQUARE,
            pxo.Property.LINEAR_NORMAL,
            pxo.Property.LINEAR_SELF_ADJOINT,
        }
        if self._cst > 0:
            preserved |= {
                pxo.Property.LINEAR_POSITIVE_DEFINITE,
                pxo.Property.QUADRATIC,
                pxo.Property.PROXIMABLE,
            }
        if self._op.has(pxo.Property.LINEAR):
            preserved.add(pxo.Property.PROXIMABLE)
        if np.isclose(self._cst, -1):
            preserved.add(pxo.Property.LINEAR_UNITARY)

        properties = self._op.properties() & preserved
        klass = pxo.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = pxu.copy_if_unsafe(self._op.apply(arr))
        out *= self._cst
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L = float(self._op.lipschitz)
        else:
            L = self._op.estimate_lipschitz(**kwargs)
        L *= abs(self._cst)
        return L

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        return self._op.prox(arr, tau * self._cst)

    def _quad_spec(self):
        Q1, c1, t1 = self._op._quad_spec()
        Q2 = ScaleRule(op=Q1, cst=self._cst).op()
        c2 = ScaleRule(op=c1, cst=self._cst).op()
        t2 = t1 * self._cst
        return (Q2, c2, t2)

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        if self.has(pxo.Property.LINEAR):
            op = self
        else:
            op = self._op.jacobian(arr) * self._cst
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            dL = float(self._op.diff_lipschitz)
        else:
            dL = self._op.estimate_diff_lipschitz(**kwargs)
        dL *= abs(self._cst)
        return dL

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = pxu.copy_if_unsafe(self._op.grad(arr))
        out *= self._cst
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = pxu.copy_if_unsafe(self._op.adjoint(arr))
        out *= self._cst
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        A = pxu.copy_if_unsafe(self._op.asarray(**kwargs))
        A *= self._cst
        return A

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = pxu.copy_if_unsafe(self._op.svdvals(**kwargs))
        D *= abs(self._cst)
        return D

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        scale = damp / (self._cst**2)
        out = pxu.copy_if_unsafe(self._op.pinv(arr, damp=scale, **kwargs))
        out /= self._cst
        return out

    def gram(self) -> pxt.OpT:
        op = self._op.gram() * (self._cst**2)
        return op

    def cogram(self) -> pxt.OpT:
        op = self._op.cogram() * (self._cst**2)
        return op

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = self._op.trace(**kwargs) * self._cst
        return tr

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        if self.has(pxo.Property.FUNCTIONAL):
            if data is None:
                op = self
            else:
                op = self._op.asloss(data) * self._cst
            return op
        else:
            raise NotImplementedError


class ArgScaleRule(Rule):
    r"""
    Arithmetic rules for element-wise parameter scaling: :math:`B(x) = A(\alpha x)`.

    Special Cases::

        \alpha = 0  => ConstantValued (w/ potential vector-valued output)
        \alpha = 1  => self

    Else::

        |--------------------------|-------------|-----------------------------------------------------------------------------|
        |         Property         |  Preserved? |                          Arithmetic Update Rule(s)                          |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | CAN_EVAL                 | yes         | op_new.apply(arr) = op_old.apply(arr * \alpha)                              |
        |                          |             | op_new.lipschitz = op_old.lipschitz * abs(\alpha)                           |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | FUNCTIONAL               | yes         | op_new.asloss(\beta) = AMBIGUOUS -> DISABLED                                |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | PROXIMABLE               | yes         | op_new.prox(arr, tau) = op_old.prox(\alpha * arr, \alpha**2 * tau) / \alpha |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | DIFFERENTIABLE           | yes         | op_new.diff_lipschitz = op_old.diff_lipschitz * (\alpha**2)                 |
        |                          |             | op_new.jacobian(arr) = op_old.jacobian(arr * \alpha) * \alpha               |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | DIFFERENTIABLE_FUNCTION  | yes         | op_new.grad(arr) = op_old.grad(\alpha * arr) * \alpha                       |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | QUADRATIC                | yes         | Q, c, t = op_old._quad_spec()                                               |
        |                          |             | op_new._quad_spec() = (\alpha**2 * Q, \alpha * c, t)                        |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR                   | yes         | op_new.adjoint(arr) = op_old.adjoint(arr) * \alpha                          |
        |                          |             | op_new.asarray() = op_old.asarray() * \alpha                                |
        |                          |             | op_new.svdvals() = op_old.svdvals() * abs(\alpha)                           |
        |                          |             | op_new.pinv(x, damp) = op_old.pinv(x, damp / (\alpha**2)) / \alpha          |
        |                          |             | op_new.gram() = op_old.gram() * (\alpha**2)                                 |
        |                          |             | op_new.cogram() = op_old.cogram() * (\alpha**2)                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_SQUARE            | yes         | op_new.trace() = op_old.trace() * \alpha                                    |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_NORMAL            | yes         |                                                                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_UNITARY           | \alpha = -1 |                                                                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_SELF_ADJOINT      | yes         |                                                                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_POSITIVE_DEFINITE | \alpha > 0  |                                                                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
        | LINEAR_IDEMPOTENT        | no          |                                                                             |
        |--------------------------|-------------|-----------------------------------------------------------------------------|
    """

    def __init__(self, op: pxt.OpT, cst: pxt.Real):
        super().__init__()
        self._op = op.squeeze()
        self._cst = float(cst)

    def op(self) -> pxt.OpT:
        if np.isclose(self._cst, 0):
            # ConstantVECTOR output: modify ConstantValued to work.
            from pyxu.operator import ConstantValued

            @pxrt.enforce_precision(i="arr")
            def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
                xp = pxu.get_array_module(arr)
                arr = xp.zeros(
                    (*arr.shape[:-1], self._op.dim),
                    dtype=arr.dtype,
                )
                out = self._op.apply(arr)
                return out

            op = ConstantValued(
                shape=self._op.shape,
                cst=self._cst,
            )
            op.apply = types.MethodType(op_apply, op)
        elif np.isclose(self._cst, 1):
            op = self._op
        else:
            klass = self._infer_op_klass()
            op = klass(shape=self._op.shape)
            op._op = self._op  # embed for introspection
            op._cst = self._cst  # embed for introspection
            for p in op.properties():
                for name in p.arithmetic_methods():
                    func = getattr(self.__class__, name)
                    setattr(op, name, types.MethodType(func, op))
            self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        return ("argscale", self._op, self._cst)

    def _infer_op_klass(self) -> pxt.OpC:
        preserved = {
            pxo.Property.CAN_EVAL,
            pxo.Property.FUNCTIONAL,
            pxo.Property.PROXIMABLE,
            pxo.Property.DIFFERENTIABLE,
            pxo.Property.DIFFERENTIABLE_FUNCTION,
            pxo.Property.LINEAR,
            pxo.Property.LINEAR_SQUARE,
            pxo.Property.LINEAR_NORMAL,
            pxo.Property.LINEAR_SELF_ADJOINT,
            pxo.Property.QUADRATIC,
        }
        if self._cst > 0:
            preserved.add(pxo.Property.LINEAR_POSITIVE_DEFINITE)
        if np.isclose(self._cst, -1):
            preserved.add(pxo.Property.LINEAR_UNITARY)

        properties = self._op.properties() & preserved
        klass = pxo.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = arr.copy()
        x *= self._cst
        out = self._op.apply(x)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L = float(self._op.lipschitz)
        else:
            L = self._op.estimate_lipschitz(**kwargs)
        L *= abs(self._cst)
        return L

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        x = arr.copy()
        x *= self._cst
        out = self._op.prox(x, (self._cst**2) * tau)
        out /= self._cst
        return out

    def _quad_spec(self):
        Q1, c1, t1 = self._op._quad_spec()
        Q2 = ScaleRule(op=Q1, cst=self._cst**2).op()
        c2 = ScaleRule(op=c1, cst=self._cst).op()
        t2 = t1
        return (Q2, c2, t2)

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        if self.has(pxo.Property.LINEAR):
            op = self
        else:
            x = arr.copy()
            x *= self._cst
            op = self._op.jacobian(x) * self._cst
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            dL = self._op.diff_lipschitz
        else:
            dL = self._op.estimate_diff_lipschitz(**kwargs)
        dL *= self._cst**2
        return dL

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = arr.copy()
        x *= self._cst
        out = self._op.grad(x)
        out *= self._cst
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = pxu.copy_if_unsafe(self._op.adjoint(arr))
        out *= self._cst
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        A = pxu.copy_if_unsafe(self._op.asarray(**kwargs))
        A *= self._cst
        return A

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = pxu.copy_if_unsafe(self._op.svdvals(**kwargs))
        D *= abs(self._cst)
        return D

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        scale = damp / (self._cst**2)
        out = pxu.copy_if_unsafe(self._op.pinv(arr, damp=scale, **kwargs))
        out /= self._cst
        return out

    def gram(self) -> pxt.OpT:
        op = self._op.gram() * (self._cst**2)
        return op

    def cogram(self) -> pxt.OpT:
        op = self._op.cogram() * (self._cst**2)
        return op

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = self._op.trace(**kwargs) * self._cst
        return tr

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        if self.has(pxo.Property.FUNCTIONAL):
            msg = "\n".join(
                [
                    "The meaning of op.argscale().asloss() is ambiguous.",
                    "Rewrite the expression differently to clarify the intent.",
                ]
            )
            raise ArithmeticError(msg)
        else:
            raise NotImplementedError


class ArgShiftRule(Rule):
    r"""
    Arithmetic rules for parameter shifting: :math:`B(x) = A(x - c)`.

    Special Cases::

        \shift = 0  => self

    Else::

        |--------------------------|------------|-----------------------------------------------------------------|
        |         Property         | Preserved? |                    Arithmetic Update Rule(s)                    |
        |--------------------------|------------|-----------------------------------------------------------------|
        | CAN_EVAL                 | yes        | op_new.apply(arr) = op_old.apply(arr + \shift)                  |
        |                          |            | op_new.lipschitz = op_old.lipschitz                             |
        |--------------------------|------------|-----------------------------------------------------------------|
        | FUNCTIONAL               | yes        | op_new.asloss(\beta) = AMBIGUOUS -> DISABLED                    |
        |--------------------------|------------|-----------------------------------------------------------------|
        | PROXIMABLE               | yes        | op_new.prox(arr, tau) = op_old.prox(arr + \shift, tau) - \shift |
        |--------------------------|------------|-----------------------------------------------------------------|
        | DIFFERENTIABLE           | yes        | op_new.diff_lipschitz = op_old.diff_lipschitz                   |
        |                          |            | op_new.jacobian(arr) = op_old.jacobian(arr + \shift)            |
        |--------------------------|------------|-----------------------------------------------------------------|
        | DIFFERENTIABLE_FUNCTION  | yes        | op_new.grad(arr) = op_old.grad(arr + \shift)                    |
        |--------------------------|------------|-----------------------------------------------------------------|
        | QUADRATIC                | yes        | Q, c, t = op_old._quad_spec()                                   |
        |                          |            | op_new._quad_spec() = (Q, c + Q @ \shift, op_old.apply(\shift)) |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR                   | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_SQUARE            | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_NORMAL            | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_UNITARY           | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_SELF_ADJOINT      | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_POSITIVE_DEFINITE | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
        | LINEAR_IDEMPOTENT        | no         |                                                                 |
        |--------------------------|------------|-----------------------------------------------------------------|
    """

    def __init__(self, op: pxt.OpT, cst: typ.Union[pxt.Real, pxt.NDArray]):
        super().__init__()
        self._op = op.squeeze()
        self._scalar = isinstance(cst, pxt.Real)
        if self._scalar:
            cst = float(cst)
        else:  # pxt.NDArray
            assert cst.size == len(cst), f"cst: expected 1D array, got {cst.shape}."
        self._cst = cst

    def op(self) -> pxt.OpT:
        kwargs = dict(fallback=np if self._scalar else None)
        xp = pxu.get_array_module(self._cst, **kwargs)
        norm = pxu.compute(xp.linalg.norm(self._cst))
        if np.isclose(float(norm), 0):
            op = self._op
        else:
            klass = self._infer_op_klass()
            shape = self._infer_op_shape()
            op = klass(shape=shape)
            op._op = self._op  # embed for introspection
            op._cst = self._cst  # embed for introspection
            for p in op.properties():
                for name in p.arithmetic_methods():
                    func = getattr(self.__class__, name)
                    setattr(op, name, types.MethodType(func, op))
            self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        sh = (None,) if isinstance(self._cst, pxt.Real) else self._cst.shape
        return ("argshift", self._op, sh)

    def _infer_op_klass(self) -> pxt.OpC:
        preserved = {
            pxo.Property.CAN_EVAL,
            pxo.Property.FUNCTIONAL,
            pxo.Property.PROXIMABLE,
            pxo.Property.DIFFERENTIABLE,
            pxo.Property.DIFFERENTIABLE_FUNCTION,
            pxo.Property.QUADRATIC,
        }

        properties = self._op.properties() & preserved
        klass = pxo.Operator._infer_operator_type(properties)
        return klass

    def _infer_op_shape(self) -> pxt.OpShape:
        if self._scalar:
            return self._op.shape
        else:  # pxt.NDArray
            dim_op = self._op.dim
            dim_cst = self._cst.size
            if (dim_op is None) or (dim_op == dim_cst):
                return (self._op.codim, dim_cst)
            else:
                raise ValueError(f"Shifting {self._op} by {self._cst.shape} forbidden.")

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = arr.copy()
        x += self._cst
        out = self._op.apply(x)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L = self._op.lipschitz
        else:
            L = self._op.estimate_lipschitz(**kwargs)
        return L

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        x = arr.copy()
        x += self._cst
        out = self._op.prox(x, tau)
        out -= self._cst
        return out

    def _quad_spec(self):
        Q1, c1, t1 = self._op._quad_spec()
        Q2 = Q1

        if isinstance(self._cst, pxt.Real):
            from pyxu.operator import Sum

            # backend-agnostic `c2`-term
            c2 = c1 + (self._cst * (Sum(arg_shape=(Q1.dim,)) * Q1))

            # Try all backends until one lets you compute `t2`.
            # (Reason: We cannot infer the backend of an operator from its public API.)
            t2 = np.nan
            for xp in pxd.supported_array_modules():
                if np.isnan(t2):
                    try:
                        cst = xp.broadcast_to(self._cst, Q1.dim)
                        t2 = float(self._op.apply(cst))
                    except Exception:
                        pass
        else:
            c2 = c1 + pxo.LinFunc.from_array(
                Q1.apply(self._cst),
                enable_warnings=False,
                # [enable_warnings] API users have no reason to call _quad_spec().
                # If they choose to use `c2`, then we assume they know what they are doing.
            )
            t2 = float(self._op.apply(self._cst))

        return (Q2, c2, t2)

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        x = arr.copy()
        x += self._cst
        op = self._op.jacobian(x)
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            dL = self._op.diff_lipschitz
        else:
            dL = self._op.estimate_diff_lipschitz(**kwargs)
        return dL

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = arr.copy()
        x += self._cst
        out = self._op.grad(x)
        return out

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        if self.has(pxo.Property.FUNCTIONAL):
            msg = "\n".join(
                [
                    "The meaning of op.argshift().asloss() is ambiguous.",
                    "Rewrite the expression differently to clarify the intent.",
                ]
            )
            raise ArithmeticError(msg)
        else:
            raise NotImplementedError


class AddRule(Rule):
    r"""
    Arithmetic rules for operator addition: :math:`C(x) = A(x) + B(x)`.

    The output type of ``AddRule(A.squeeze(), B.squeeze())`` is summarized in the table below (LHS/RHS
    commute)::

        |---------------|-----|------|---------|----------|----------|--------------|-----------|---------|--------------|------------|------------|------------|---------------|---------------|------------|---------------|
        |   LHS / RHS   | Map | Func | DiffMap | DiffFunc | ProxFunc | ProxDiffFunc | Quadratic |  LinOp  |   LinFunc    |  SquareOp  |  NormalOp  |   UnitOp   | SelfAdjointOp |    PosDefOp   |   ProjOp   |   OrthProjOp  |
        |---------------|-----|------|---------|----------|----------|--------------|-----------|---------|--------------|------------|------------|------------|---------------|---------------|------------|---------------|
        | Map           | Map | Map  | Map     | Map      | Map      | Map          | Map       | Map     | Map          | Map        | Map        | Map        | Map           | Map           | Map        | Map           |
        | Func          |     | Func | Map     | Func     | Func     | Func         | Func      | Map     | Func         | Map        | Map        | Map        | Map           | Map           | Map        | Map           |
        | DiffMap       |     |      | DiffMap | DiffMap  | Map      | DiffMap      | DiffMap   | DiffMap | DiffMap      | DiffMap    | DiffMap    | DiffMap    | DiffMap       | DiffMap       | DiffMap    | DiffMap       |
        | DiffFunc      |     |      |         | DiffFunc | Func     | DiffFunc     | DiffFunc  | DiffMap | DiffFunc     | DiffMap    | DiffMap    | DiffMap    | DiffMap       | DiffMap       | DiffMap    | DiffMap       |
        | ProxFunc      |     |      |         |          | Func     | Func         | Func      | Map     | ProxFunc     | Map        | Map        | Map        | Map           | Map           | Map        | Map           |
        | ProxDiffFunc  |     |      |         |          |          | DiffFunc     | DiffFunc  | DiffMap | ProxDiffFunc | DiffMap    | DiffMap    | DiffMap    | DiffMap       | DiffMap       | DiffMap    | DiffMap       |
        | Quadratic     |     |      |         |          |          |              | Quadratic | DiffMap | Quadratic    | DiffMap    | DiffMap    | DiffMap    | DiffMap       | DiffMap       | DiffMap    | DiffMap       |
        | LinOp         |     |      |         |          |          |              |           | LinOp   | LinOp        | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE    | IMPOSSIBLE    | IMPOSSIBLE | IMPOSSIBLE    |
        | LinFunc       |     |      |         |          |          |              |           |         | LinFunc      | SquareOp   | SquareOp   | SquareOp   | SquareOp      | SquareOp      | SquareOp   | SquareOp      |
        | SquareOp      |     |      |         |          |          |              |           |         |              | SquareOp   | SquareOp   | SquareOp   | SquareOp      | SquareOp      | SquareOp   | SquareOp      |
        | NormalOp      |     |      |         |          |          |              |           |         |              |            | SquareOp   | SquareOp   | SquareOp      | SquareOp      | SquareOp   | SquareOp      |
        | UnitOp        |     |      |         |          |          |              |           |         |              |            |            | SquareOp   | SquareOp      | SquareOp      | SquareOp   | SquareOp      |
        | SelfAdjointOp |     |      |         |          |          |              |           |         |              |            |            |            | SelfAdjointOp | SelfAdjointOp | SquareOp   | SelfAdjointOp |
        | PosDefOp      |     |      |         |          |          |              |           |         |              |            |            |            |               | PosDefOp      | SquareOp   | PosDefOp      |
        | ProjOp        |     |      |         |          |          |              |           |         |              |            |            |            |               |               | SquareOp   | SquareOp      |
        | OrthProjOp    |     |      |         |          |          |              |           |         |              |            |            |            |               |               |            | SelfAdjointOp |
        |---------------|-----|------|---------|----------|----------|--------------|-----------|---------|--------------|------------|------------|------------|---------------|---------------|------------|---------------|

    Arithmetic Update Rule(s)::

        * CAN_EVAL
            op.apply(arr) = _lhs.apply(arr) + _rhs.apply(arr)
            op.lipschitz = _lhs.lipschitz + _rhs.lipschitz
            IMPORTANT: if range-broadcasting takes place (ex: LHS(1,) + RHS(M,)), then the broadcasted
                       operand's Lipschitz constant must be magnified by \sqrt{M}.

        * FUNCTIONAL
            op.asloss(\beta) = _lhs.asloss(\beta) + _rhs.asloss(\beta)
                               may be ambiguous -> warning

        * PROXIMABLE
            op.prox(arr, tau) = _lhs.prox(arr - tau * _rhs.grad(arr), tau)
                          OR  = _rhs.prox(arr - tau * _lhs.grad(arr), tau)
                IMPORTANT: the one calling .grad() should be either (lhs, rhs) which has LINEAR property

        * DIFFERENTIABLE
            op.jacobian(arr) = _lhs.jacobian(arr) + _rhs.jacobian(arr)
            op.diff_lipschitz = _lhs.diff_lipschitz + _rhs.diff_lipschitz
            IMPORTANT: if range-broadcasting takes place (ex: LHS(1,) + RHS(M,)), then the broadcasted
                       operand's diff-Lipschitz constant must be magnified by \sqrt{M}.

        * DIFFERENTIABLE_FUNCTION
            op.grad(arr) = _lhs.grad(arr) + _rhs.grad(arr)

        * LINEAR
            op.adjoint(arr) = _lhs.adjoint(arr) + _rhs.adjoint(arr)
            IMPORTANT: if range-broadcasting takes place (ex: LHS(1,) + RHS(M,)), then the broadcasted
                       operand's adjoint-input must be averaged.
            op.asarray() = _lhs.asarray() + _rhs.asarray()
            op.gram() = _lhs.gram() + _rhs.gram() + (_lhs.T * _rhs) + (_rhs.T * _lhs)
            op.cogram() = _lhs.cogram() + _rhs.cogram() + (_lhs * _rhs.T) + (_rhs * _lhs.T)

        * LINEAR_SQUARE
            op.trace() = _lhs.trace() + _rhs.trace()

        * QUADRATIC
            lhs = rhs = quadratic
              Q_l, c_l, t_l = lhs._quad_spec()
              Q_r, c_r, t_r = rhs._quad_spec()
              op._quad_spec() = (Q_l + Q_r, c_l + c_r, t_l + t_r)
            lhs, rhs = quadratic, linear
              Q, c, t = lhs._quad_spec()
              op._quad_spec() = (Q, c + rhs, t)
    """

    def __init__(self, lhs: pxt.OpT, rhs: pxt.OpT):
        super().__init__()
        self._lhs = lhs.squeeze()
        self._rhs = rhs.squeeze()

    def op(self) -> pxt.OpT:
        sh_op = pxu.infer_sum_shape(self._lhs.shape, self._rhs.shape)
        klass = self._infer_op_klass()
        if klass.has(pxo.Property.QUADRATIC):
            # Quadratic additive arithmetic differs substantially from other arithmetic operations.
            # To avoid tedious redefinitions of arithmetic methods to handle QuadraticFunc
            # specifically, the code-path below delegates additive arithmetic directly to
            # QuadraticFunc.
            lin = lambda _: _.has(pxo.Property.LINEAR)
            quad = lambda _: _.has(pxo.Property.QUADRATIC)

            if quad(self._lhs) and quad(self._rhs):
                lQ, lc, lt = self._lhs._quad_spec()
                rQ, rc, rt = self._rhs._quad_spec()
                op = klass(
                    shape=sh_op,
                    Q=lQ + rQ,
                    c=lc + rc,
                    t=lt + rt,
                )
            elif quad(self._lhs) and lin(self._rhs):
                lQ, lc, lt = self._lhs._quad_spec()
                op = klass(
                    shape=sh_op,
                    Q=lQ,
                    c=lc + self._rhs,
                    t=lt,
                )
            elif lin(self._lhs) and quad(self._rhs):
                rQ, rc, rt = self._rhs._quad_spec()
                op = klass(
                    shape=sh_op,
                    Q=rQ,
                    c=self._lhs + rc,
                    t=rt,
                )
            else:
                raise ValueError("Impossible scenario: something went wrong during klass inference.")
        else:
            op = klass(shape=sh_op)
            op._lhs = self._lhs  # embed for introspection
            op._rhs = self._rhs  # embed for introspection
            for p in op.properties():
                for name in p.arithmetic_methods():
                    func = getattr(self.__class__, name)
                    setattr(op, name, types.MethodType(func, op))
            self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        return ("add", self._lhs, self._rhs)

    def _infer_op_klass(self) -> pxt.OpC:
        P = pxo.Property
        lhs_p = self._lhs.properties()
        rhs_p = self._rhs.properties()
        base = set(lhs_p & rhs_p)
        base.discard(pxo.Property.LINEAR_NORMAL)
        base.discard(pxo.Property.LINEAR_UNITARY)
        base.discard(pxo.Property.LINEAR_IDEMPOTENT)
        base.discard(pxo.Property.PROXIMABLE)

        # Exceptions ----------------------------------------------------------
        # normality preserved for self-adjoint addition
        if P.LINEAR_SELF_ADJOINT in base:
            base.add(P.LINEAR_NORMAL)

        # orth-proj + pos-def => pos-def
        if (({P.LINEAR_IDEMPOTENT, P.LINEAR_SELF_ADJOINT} < lhs_p) and (P.LINEAR_POSITIVE_DEFINITE in rhs_p)) or (
            ({P.LINEAR_IDEMPOTENT, P.LINEAR_SELF_ADJOINT} < rhs_p) and (P.LINEAR_POSITIVE_DEFINITE in lhs_p)
        ):
            base.add(P.LINEAR_SQUARE)
            base.add(P.LINEAR_NORMAL)
            base.add(P.LINEAR_SELF_ADJOINT)
            base.add(P.LINEAR_POSITIVE_DEFINITE)

        # linfunc + (square-shape) => square
        if P.LINEAR in base:
            sh = pxu.infer_sum_shape(self._lhs.shape, self._rhs.shape)
            if (sh[0] == sh[1]) and (sh[0] > 1):
                base.add(P.LINEAR_SQUARE)

        # quadratic + quadratic => quadratic
        if P.QUADRATIC in base:
            base.add(P.PROXIMABLE)

        # quadratic + linfunc => quadratic
        if (P.PROXIMABLE in (lhs_p & rhs_p)) and ({P.QUADRATIC, P.LINEAR} < (lhs_p | rhs_p)):
            base.add(P.QUADRATIC)

        # prox(-diff) + linfunc => prox(-diff)
        if (P.PROXIMABLE in (lhs_p & rhs_p)) and (P.LINEAR in (lhs_p | rhs_p)):
            base.add(P.PROXIMABLE)

        klass = pxo.Operator._infer_operator_type(base)
        return klass

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        # ranges may broadcast, so can't do in-place updates.
        out_lhs = self._lhs.apply(arr)
        out_rhs = self._rhs.apply(arr)
        out = out_lhs + out_rhs
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L_lhs = self._lhs.lipschitz
            L_rhs = self._rhs.lipschitz
        elif self.has(pxo.Property.LINEAR):
            L = self.__class__.estimate_lipschitz(self, **kwargs)
            return L
        else:
            L_lhs = self._lhs.estimate_lipschitz(**kwargs)
            L_rhs = self._rhs.estimate_lipschitz(**kwargs)

        # Account for broadcasting
        if self._lhs.codim < self._rhs.codim:
            # LHS broadcasts
            L_lhs = L_lhs * np.sqrt(self._rhs.codim)
        elif self._lhs.codim > self._rhs.codim:
            # RHS broadcasts
            L_rhs = L_rhs * np.sqrt(self._lhs.codim)

        L = L_lhs + L_rhs
        return L

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        P_LHS = self._lhs.properties()
        P_RHS = self._rhs.properties()
        if pxo.Property.LINEAR in (P_LHS | P_RHS):
            # linear + proximable
            if pxo.Property.LINEAR in P_LHS:
                P, G = self._rhs, self._lhs
            elif pxo.Property.LINEAR in P_RHS:
                P, G = self._lhs, self._rhs
            x = pxu.copy_if_unsafe(G.grad(arr))
            x *= -tau
            x += arr
            out = P.prox(x, tau)
        else:
            raise NotImplementedError
        return out

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        if self.has(pxo.Property.LINEAR):
            op = self
        else:
            op_lhs = self._lhs.jacobian(arr)
            op_rhs = self._rhs.jacobian(arr)
            op = op_lhs + op_rhs
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            dL_lhs = self._lhs.diff_lipschitz
            dL_rhs = self._rhs.diff_lipschitz
        elif self.has(pxo.Property.LINEAR):
            dL_lhs = 0
            dL_rhs = 0
        else:
            dL_lhs = self._lhs.estimate_diff_lipschitz(**kwargs)
            dL_rhs = self._rhs.estimate_diff_lipschitz(**kwargs)

        # Account for broadcasting
        if self._lhs.codim < self._rhs.codim:
            # LHS broadcasts
            dL_lhs = dL_lhs * np.sqrt(self._rhs.codim)
        elif self._lhs.codim > self._rhs.codim:
            # RHS broadcasts
            dL_rhs = dL_rhs * np.sqrt(self._lhs.codim)

        dL = dL_lhs + dL_rhs
        return dL

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = self._lhs.grad(arr)
        out = pxu.copy_if_unsafe(out)
        out += self._rhs.grad(arr)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        arr_lhs = arr_rhs = arr
        if self._lhs.codim < self._rhs.codim:
            # LHS broadcasts
            arr_lhs = arr.sum(axis=-1, keepdims=True)
        elif self._lhs.codim > self._rhs.codim:
            # RHS broadcasts
            arr_rhs = arr.sum(axis=-1, keepdims=True)

        out = self._lhs.adjoint(arr_lhs)
        out = pxu.copy_if_unsafe(out)
        out += self._rhs.adjoint(arr_rhs)
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        # broadcast may be involved, so can't do in-place updates.
        A_lhs = self._lhs.asarray(**kwargs)
        A_rhs = self._rhs.asarray(**kwargs)
        A = A_lhs + A_rhs
        return A

    def gram(self) -> pxt.OpT:
        lhs, rhs = self._lhs, self._rhs
        if self._lhs.codim == self._rhs.codim:
            # No broadcasting
            lhs_F = rhs_F = False
        else:
            # Broadcasting
            lhs_F = self._lhs.has(pxo.Property.FUNCTIONAL)
            rhs_F = self._rhs.has(pxo.Property.FUNCTIONAL)
            if lhs_F:
                lhs = _Sum(M=self._rhs.codim, N=1) * self._lhs
            if rhs_F:
                rhs = _Sum(M=self._lhs.codim, N=1) * self._rhs

        op1 = self._lhs.gram() * (self._rhs.codim if lhs_F else 1)
        op2 = self._rhs.gram() * (self._lhs.codim if rhs_F else 1)
        op3 = lhs.T * rhs
        op4 = rhs.T * lhs
        op = op1 + op2 + (op3 + op4).asop(pxo.SelfAdjointOp)
        return op.squeeze()

    def cogram(self) -> pxt.OpT:
        lhs, rhs = self._lhs, self._rhs
        if self._lhs.codim == self._rhs.codim:
            # No broadcasting
            lhs_F = rhs_F = False
        else:
            # Broadcasting
            lhs_F = self._lhs.has(pxo.Property.FUNCTIONAL)
            rhs_F = self._rhs.has(pxo.Property.FUNCTIONAL)
            if lhs_F:
                lhs = _Sum(M=self._rhs.codim, N=1) * self._lhs
            if rhs_F:
                rhs = _Sum(M=self._lhs.codim, N=1) * self._rhs

        if lhs_F:
            scale = float(self._lhs.cogram().asarray())
            op1 = _Sum(M=self._rhs.codim, N=self._rhs.codim) * scale
        else:
            op1 = self._lhs.cogram()
        if rhs_F:
            scale = float(self._rhs.cogram().asarray())
            op2 = _Sum(M=self._lhs.codim, N=self._lhs.codim) * scale
        else:
            op2 = self._rhs.cogram()
        op3 = lhs * rhs.T
        op4 = rhs * lhs.T
        op = op1 + op2 + (op3 + op4).asop(pxo.SelfAdjointOp)
        return op.squeeze()

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = 0
        for side in (self._lhs, self._rhs):
            if side.has(pxo.Property.FUNCTIONAL):
                tr += float(side.asarray().sum())
            else:
                tr += float(side.trace(**kwargs))
        return tr

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        if self.has(pxo.Property.FUNCTIONAL):
            msg = "\n".join(
                [
                    "The meaning of (lhs + rhs).asloss() may be ambiguous if the loss-notion differs among functionals involved.",
                    "It is recommended to call asloss() prior to adding functionals instead:",
                    "    lhs.asloss(data) + rhs.asloss(data)",
                ]
            )
            warnings.warn(msg, pxw.AutoInferenceWarning)

            if data is None:
                op = self
            else:
                op_lhs = self._lhs.asloss(data=data)
                op_rhs = self._rhs.asloss(data=data)
                op = op_lhs + op_rhs
            return op
        else:
            raise NotImplementedError


class ChainRule(Rule):
    r"""
    Arithmetic rules for operator composition: :math:`C(x) = (A \circ B)(x)`.

    The output type of ``ChainRule(A.squeeze(), B.squeeze())`` is summarized in the table below::

        |---------------|------|------------|----------|------------|------------|----------------|----------------------|------------------|------------|-----------|-----------|--------------|---------------|-----------|-----------|------------|
        |   LHS / RHS   | Map  |    Func    | DiffMap  |  DiffFunc  |  ProxFunc  |  ProxDiffFunc  |      Quadratic       |      LinOp       |  LinFunc   |  SquareOp |  NormalOp |    UnitOp    | SelfAdjointOp |  PosDefOp |   ProjOp  | OrthProjOp |
        |---------------|------|------------|----------|------------|------------|----------------|----------------------|------------------|------------|-----------|-----------|--------------|---------------|-----------|-----------|------------|
        | Map           | Map  | Map        | Map      | Map        | Map        | Map            | Map                  | Map              | Map        | Map       | Map       | Map          | Map           | Map       | Map       | Map        |
        | Func          | Func | Func       | Func     | Func       | Func       | Func           | Func                 | Func             | Func       | Func      | Func      | Func         | Func          | Func      | Func      | Func       |
        | DiffMap       | Map  | Map        | DiffMap  | DiffMap    | Map        | DiffMap        | DiffMap              | DiffMap          | DiffMap    | DiffMap   | DiffMap   | DiffMap      | DiffMap       | DiffMap   | DiffMap   | DiffMap    |
        | DiffFunc      | Func | Func       | DiffFunc | DiffFunc   | Func       | DiffFunc       | DiffFunc             | DiffFunc         | DiffFunc   | DiffFunc  | DiffFunc  | DiffFunc     | DiffFunc      | DiffFunc  | DiffFunc  | DiffFunc   |
        | ProxFunc      | Func | Func       | Func     | Func       | Func       | Func           | Func                 | Func             | Func       | Func      | Func      | ProxFunc     | Func          | Func      | Func      | Func       |
        | ProxDiffFunc  | Func | Func       | DiffFunc | DiffFunc   | Func       | DiffFunc       | DiffFunc             | DiffFunc         | DiffFunc   | DiffFunc  | DiffFunc  | ProxDiffFunc | DiffFunc      | DiffFunc  | DiffFunc  | DiffFunc   |
        | Quadratic     | Func | Func       | DiffFunc | DiffFunc   | Func       | DiffFunc       | DiffFunc             | Quadratic        | Quadratic  | Quadratic | Quadratic | Quadratic    | Quadratic     | Quadratic | Quadratic | Quadratic  |
        | LinOp         | Map  | Func       | DiffMap  | DiffMap    | Map        | DiffMap        | DiffMap              | LinOp / SquareOp | LinOp      | LinOp     | LinOp     | LinOp        | LinOp         | LinOp     | LinOp     | LinOp      |
        | LinFunc       | Func | Func       | DiffFunc | DiffFunc   | [Prox]Func | [Prox]DiffFunc | DiffFunc / Quadratic | LinFunc          | LinFunc    | LinFunc   | LinFunc   | LinFunc      | LinFunc       | LinFunc   | LinFunc   | LinFunc    |
        | SquareOp      | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | NormalOp      | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | UnitOp        | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | UnitOp       | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | SelfAdjointOp | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | PosDefOp      | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | ProjOp        | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        | OrthProjOp    | Map  | IMPOSSIBLE | DiffMap  | IMPOSSIBLE | IMPOSSIBLE | IMPOSSIBLE     | IMPOSSIBLE           | LinOp            | IMPOSSIBLE | SquareOp  | SquareOp  | SquareOp     | SquareOp      | SquareOp  | SquareOp  | SquareOp   |
        |---------------|------|------------|----------|------------|------------|----------------|----------------------|------------------|------------|-----------|-----------|--------------|---------------|-----------|-----------|------------|

    Arithmetic Update Rule(s)::

        * CAN_EVAL
            op.apply(arr) = _lhs.apply(_rhs.apply(arr))
            op.lipschitz = _lhs.lipschitz * _rhs.lipschitz

        * FUNCTIONAL
            op.asloss(\beta) = ambiguous -> disabled

        * PROXIMABLE (RHS Unitary only)
            op.prox(arr, tau) = _rhs.adjoint(_lhs.prox(_rhs.apply(arr), tau))

        * DIFFERENTIABLE
            op.jacobian(arr) = _lhs.jacobian(_rhs.apply(arr)) * _rhs.jacobian(arr)
            op.diff_lipschitz =
                quadratic            => _quad_spec().Q.lipschitz
                linear \comp linear  => 0
                linear \comp diff    => _lhs.lipschitz * _rhs.diff_lipschitz
                diff   \comp linear  => _lhs.diff_lipschitz * (_rhs.lipschitz ** 2)
                diff   \comp diff    => \infty

        * DIFFERENTIABLE_FUNCTION (1D input)
            op.grad(arr) = _lhs.grad(_rhs.apply(arr)) @ _rhs.jacobian(arr).asarray()

        * LINEAR
            op.adjoint(arr) = _rhs.adjoint(_lhs.adjoint(arr))
            op.asarray() = _lhs.asarray() @ _rhs.asarray()
            op.gram() = _rhs.T @ _lhs.gram() @ _rhs
            op.cogram() = _lhs @ _rhs.cogram() @ _lhs.T

        * QUADRATIC
            Q, c, t = _lhs._quad_spec()
            op._quad_spec() = (_rhs.T * Q * _rhs, _rhs.T * c, t)
    """

    def __init__(self, lhs: pxt.OpT, rhs: pxt.OpT):
        super().__init__()
        self._lhs = lhs.squeeze()
        self._rhs = rhs.squeeze()

    def op(self) -> pxt.OpT:
        sh_op = pxu.infer_composition_shape(self._lhs.shape, self._rhs.shape)
        klass = self._infer_op_klass()
        op = klass(shape=sh_op)
        op._lhs = self._lhs  # embed for introspection
        op._rhs = self._rhs  # embed for introspection
        for p in op.properties():
            for name in p.arithmetic_methods():
                func = getattr(self.__class__, name)
                setattr(op, name, types.MethodType(func, op))
        self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        return ("compose", self._lhs, self._rhs)

    def _infer_op_klass(self) -> pxt.OpC:
        # |--------------------------|------------------------------------------------------|
        # |         Property         |                      Preserved?                      |
        # |--------------------------|------------------------------------------------------|
        # | CAN_EVAL                 | (LHS CAN_EVAL) & (RHS CAN_EVAL)                      |
        # |--------------------------|------------------------------------------------------|
        # | FUNCTIONAL               | LHS FUNCTIONAL                                       |
        # |--------------------------|------------------------------------------------------|
        # | PROXIMABLE               | * (LHS PROXIMABLE) & (RHS LINEAR_UNITARY)            |
        # |                          | * (LHS LINEAR) & (RHS LINEAR)                        |
        # |                          | * (LHS LINEAR FUNCTIONAL [> 0]) & (RHS PROXIMABLE)   |
        # |--------------------------|------------------------------------------------------|
        # | DIFFERENTIABLE           | (LHS DIFFERENTIABLE) & (RHS DIFFERENTIABLE)          |
        # |--------------------------|------------------------------------------------------|
        # | DIFFERENTIABLE_FUNCTION  | (LHS DIFFERENTIABLE_FUNCTION) & (RHS DIFFERENTIABLE) |
        # |--------------------------|------------------------------------------------------|
        # | QUADRATIC                | * (LHS QUADRATIC) & (RHS LINEAR)                     |
        # |                          | * (LHS LINEAR FUNCTIONAL [> 0]) & (RHS QUADRATIC)    |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR                   | (LHS LINEAR) & (RHS LINEAR)                          |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_SQUARE            | (Shape[LHS * RHS] square) & (LHS.codim > 1)          |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_NORMAL            | no                                                   |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_UNITARY           | (LHS LINEAR_UNITARY) & (RHS LINEAR_UNITARY)          |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_SELF_ADJOINT      | no                                                   |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_POSITIVE_DEFINITE | no                                                   |
        # |--------------------------|------------------------------------------------------|
        # | LINEAR_IDEMPOTENT        | no                                                   |
        # |--------------------------|------------------------------------------------------|
        lhs_p = self._lhs.properties()
        rhs_p = self._rhs.properties()
        P = pxo.Property
        properties = {P.CAN_EVAL}
        if P.FUNCTIONAL in lhs_p:
            properties.add(P.FUNCTIONAL)
        # Proximal ------------------------------
        if (P.PROXIMABLE in lhs_p) and (P.LINEAR_UNITARY in rhs_p):
            properties.add(P.PROXIMABLE)
        elif ({P.LINEAR, P.FUNCTIONAL} < lhs_p) and (P.PROXIMABLE in rhs_p):
            cst = self._lhs.asarray().item()
            if cst > 0:
                properties.add(P.PROXIMABLE)
                if P.QUADRATIC in rhs_p:
                    properties.add(P.QUADRATIC)
        # ---------------------------------------
        if P.DIFFERENTIABLE in (lhs_p & rhs_p):
            properties.add(P.DIFFERENTIABLE)
        if (P.DIFFERENTIABLE_FUNCTION in lhs_p) and (P.DIFFERENTIABLE in rhs_p):
            properties.add(P.DIFFERENTIABLE_FUNCTION)
        if (P.QUADRATIC in lhs_p) and (P.LINEAR in rhs_p):
            properties.add(P.PROXIMABLE)
            properties.add(P.QUADRATIC)
        if P.LINEAR in (lhs_p & rhs_p):
            properties.add(P.LINEAR)
            if self._lhs.codim == 1:
                properties.add(P.PROXIMABLE)
            if self._lhs.codim == self._rhs.dim > 1:
                properties.add(P.LINEAR_SQUARE)
        if P.LINEAR_UNITARY in (lhs_p & rhs_p):
            properties.add(P.LINEAR_NORMAL)
            properties.add(P.LINEAR_UNITARY)

        klass = pxo.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = self._rhs.apply(arr)
        out = self._lhs.apply(x)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L_lhs = self._lhs.lipschitz
            L_rhs = self._rhs.lipschitz
        elif self.has(pxo.Property.LINEAR):
            L = self.__class__.estimate_lipschitz(self, **kwargs)
            return L
        else:
            L_lhs = self._lhs.estimate_lipschitz(**kwargs)
            L_rhs = self._rhs.estimate_lipschitz(**kwargs)

        zeroQ = lambda _: np.isclose(_, 0)
        if zeroQ(L_lhs) or zeroQ(L_rhs):
            L = 0
        else:
            L = L_lhs * L_rhs
        return L

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        if self.has(pxo.Property.PROXIMABLE):
            out = None
            if self._lhs.has(pxo.Property.PROXIMABLE) and self._rhs.has(pxo.Property.LINEAR_UNITARY):
                # prox[diff]func() \comp unitop() => prox[diff]func()
                x = self._rhs.apply(arr)
                y = self._lhs.prox(x, tau)
                out = self._rhs.adjoint(y)
            elif self._lhs.has(pxo.Property.QUADRATIC) and self._rhs.has(pxo.Property.LINEAR):
                # quadratic \comp linop => quadratic
                Q, c, t = self._quad_spec()
                out = pxo.QuadraticFunc(shape=self.shape, Q=Q, c=c, t=t).prox(arr, tau)
            elif self._lhs.has(pxo.Property.LINEAR) and self._rhs.has(pxo.Property.PROXIMABLE):
                # linfunc() \comp prox[diff]func() => prox[diff]func()
                #                                  = (\alpha * prox[diff]func())
                op = ScaleRule(op=self._rhs, cst=self._lhs.asarray().item()).op()
                out = op.prox(arr, tau)
            elif pxo.Property.LINEAR in (self._lhs.properties() & self._rhs.properties()):
                # linfunc() \comp linop() => linfunc()
                out = pxo.LinFunc.prox(self, arr, tau)

            if out is not None:
                return out
        raise NotImplementedError

    def _quad_spec(self):
        if self.has(pxo.Property.QUADRATIC):
            if self._lhs.has(pxo.Property.LINEAR):
                # linfunc (scalar) \comp quadratic
                op = ScaleRule(op=self._rhs, cst=self._lhs.asarray().item()).op()
                Q2, c2, t2 = op._quad_spec()
            elif self._rhs.has(pxo.Property.LINEAR):
                # quadratic \comp linop
                Q1, c1, t1 = self._lhs._quad_spec()
                Q2 = (self._rhs.T * Q1 * self._rhs).asop(pxo.PosDefOp)
                c2 = c1 * self._rhs
                t2 = t1
            return (Q2, c2, t2)
        else:
            raise NotImplementedError

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        if self.has(pxo.Property.LINEAR):
            op = self
        else:
            J_rhs = self._rhs.jacobian(arr)
            J_lhs = self._lhs.jacobian(self._rhs.apply(arr))
            op = J_lhs * J_rhs
        return op

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if self.has(pxo.Property.QUADRATIC):
            Q, c, t = self._quad_spec()
            op = pxo.QuadraticFunc(shape=self.shape, Q=Q, c=c, t=t)
            if no_eval:
                dL = op.diff_lipschitz
            else:
                dL = op.estimate_diff_lipschitz(**kwargs)
        elif self._lhs.has(pxo.Property.LINEAR) and self._rhs.has(pxo.Property.LINEAR):
            dL = 0
        elif self._lhs.has(pxo.Property.LINEAR) and self._rhs.has(pxo.Property.DIFFERENTIABLE):
            if no_eval:
                L_lhs = self._lhs.lipschitz
                dL_rhs = self._rhs.diff_lipschitz
            else:
                L_lhs = self._lhs.estimate_lipschitz(**kwargs)
                dL_rhs = self._rhs.estimate_diff_lipschitz(**kwargs)
            dL = L_lhs * dL_rhs
        elif self._lhs.has(pxo.Property.DIFFERENTIABLE) and self._rhs.has(pxo.Property.LINEAR):
            if no_eval:
                dL_lhs = self._lhs.diff_lipschitz
                L_rhs = self._rhs.lipschitz
            else:
                dL_lhs = self._lhs.estimate_diff_lipschitz(**kwargs)
                L_rhs = self._rhs.estimate_lipschitz(**kwargs)
            dL = dL_lhs * (L_rhs**2)
        else:
            dL = np.inf
        return dL

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = self._lhs.grad(self._rhs.apply(arr))
        if (arr.ndim == 1) or self._rhs.has(pxo.Property.LINEAR):
            out = self._rhs.jacobian(arr).adjoint(x)
        else:
            xp = pxu.get_array_module(arr)
            out = xp.stack(
                [
                    self._rhs.jacobian(a).adjoint(b)
                    for (a, b) in zip(
                        arr.reshape((np.prod(arr.shape[:-1]), -1)),
                        x.reshape((np.prod(x.shape[:-1]), -1)),
                        # zip() above safer than
                        #       zip(
                        #         arr.reshape((-1, self._rhs.dim)),
                        #         x.reshape((-1, self._lhs.dim)),
                        #       )
                        # Due to potential _lhs/_rhs domain-agnosticity, i.e. `[lhs|rhs].dim=None`.
                    )
                ],
                axis=0,
            ).reshape(arr.shape)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = self._lhs.adjoint(arr)
        out = self._rhs.adjoint(x)
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        A_lhs = self._lhs.asarray(**kwargs)
        A_rhs = self._rhs.asarray(**kwargs)
        A = A_lhs @ A_rhs
        return A

    def gram(self) -> pxt.OpT:
        op = self._rhs.T * self._lhs.gram() * self._rhs
        return op.asop(pxo.SelfAdjointOp).squeeze()

    def cogram(self) -> pxt.OpT:
        op = self._lhs * self._rhs.cogram() * self._lhs.T
        return op.asop(pxo.SelfAdjointOp).squeeze()

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        if self.has(pxo.Property.FUNCTIONAL):
            msg = "\n".join(
                [
                    "The meaning of (lhs * rhs).asloss() is ambiguous:",
                    "    (1) (lhs * rhs).asloss(data) ?= lhs.asloss(data) * rhs",
                    "    (2) (lhs * rhs).asloss(data) ?= lhs * g.[unknown_transform](data)",
                    "Rewrite the expression differently to clarify the intent.",
                ]
            )
            raise ArithmeticError(msg)
        else:
            raise NotImplementedError


class PowerRule(Rule):
    r"""
    Arithmetic rules for operator exponentiation: :math:`B(x) = A^{k}(x)`.

    Special Cases::

        k = 0  => IdentityOp

    Else::

        B = A \circ ... \circ A  (k-1 compositions)
    """

    def __init__(self, op: pxt.OpT, k: pxt.Integer):
        super().__init__()
        assert op.codim == op.dim, f"PowerRule: expected endomorphism, got {op}."
        assert int(k) >= 0, "PowerRule: only non-negative exponents are supported."
        self._op = op.squeeze()
        self._k = int(k)

    def op(self) -> pxt.OpT:
        if self._k == 0:
            from pyxu.operator import IdentityOp

            op = IdentityOp(dim=self._op.codim)
        else:
            op = self._op
            if pxo.Property.LINEAR_IDEMPOTENT not in self._op.properties():
                for _ in range(self._k - 1):
                    op = ChainRule(self._op, op).op()

                # Needed due to implicit PowerRule definition in terms of ChainRule.
                op._expr = self._expr
            else:
                # To stop .expr() from recursing indefinitely.
                op._expr = self._op._expr
        return op

    def _expr(self) -> tuple:
        return ("exp", self._op, self._k)


class TransposeRule(Rule):
    # Not strictly-speaking an arithmetic method, but the logic behind constructing transposed
    # operators is identical to arithmetic methods.
    # LinOp.T() rules are hence summarized here.
    r"""
    Arithmetic rules for :py:class:`~pyxu.abc.LinOp` transposition: :math:`B(x) = A^{T}(x)`.

    Arithmetic Update Rule(s)::

        * CAN_EVAL
            opT.apply(arr) = op.adjoint(arr)
            opT.lipschitz = op.lipschitz

        * FUNCTIONAL
            opT.asloss(\beta) = UNDEFINED

        * PROXIMABLE
            opT.prox(arr, tau) = LinFunc.prox(arr, tau)

        * DIFFERENTIABLE
            opT.jacobian(arr) = opT
            opT.diff_lipschitz = 0

        * DIFFERENTIABLE_FUNCTION
            opT.grad(arr) = LinFunc.grad(arr)

        * LINEAR
            opT.adjoint(arr) = op.apply(arr)
            opT.asarray() = op.asarray().T
            opT.gram() = op.cogram()
            opT.cogram() = op.gram()
            opT.svdvals() = op.svdvals()

        * LINEAR_SQUARE
            opT.trace() = op.trace()
    """

    def __init__(self, op: pxt.OpT):
        super().__init__()
        self._op = op

    def op(self) -> pxt.OpT:
        klass = self._infer_op_klass()
        op = klass(shape=(self._op.dim, self._op.codim))
        op._op = self._op  # embed for introspection
        for p in op.properties():
            for name in p.arithmetic_methods():
                func = getattr(self.__class__, name)
                setattr(op, name, types.MethodType(func, op))
        self._propagate_constants(op)
        return op

    def _expr(self) -> tuple:
        return ("transpose", self._op)

    def _infer_op_klass(self) -> pxt.OpC:
        # |----------------------|-----------------------|
        # | op_klass(codim, dim) | opT_klass(codim, dim) |
        # |----------------------|-----------------------|
        # | LinFunc(1, N)        | LinOp(N, 1)           |
        # | LinOp(N, 1)          | LinFunc(1, N)         |
        # | SquareOp(N, N)       | SquareOp(N, N)        |
        # |----------------------|-----------------------|
        properties = self._op.properties()
        if self._op.codim == self._op.dim == 1:
            klass = pxo.LinFunc
        elif pxo.Property.FUNCTIONAL in properties:
            klass = pxo.LinOp
        elif self._op.dim == 1:
            klass = pxo.LinFunc
        else:
            klass = pxo.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = self._op.adjoint(arr)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L = self._op.lipschitz
        else:
            L = self._op.estimate_lipschitz(**kwargs)
        return L

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        raise NotImplementedError

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        out = pxo.LinFunc.prox(self, arr, tau)
        return out

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        return self

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        return 0

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = pxo.LinFunc.grad(self, arr)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = self._op.apply(arr)
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        A = self._op.asarray(**kwargs)
        return A.T

    def gram(self) -> pxt.OpT:
        op = self._op.cogram()
        return op

    def cogram(self) -> pxt.OpT:
        op = self._op.gram()
        return op

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = self._op.svdvals(**kwargs)
        return D

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = self._op.trace(**kwargs)
        return tr


# Helper Class/Functions ------------------------------------------------------
def _Sum(M: int, N: int) -> pxt.OpT:
    # f: \bR^{N} -> \bR^{M}
    #      x     -> [sum(x), ..., sum(x)]  (M times)
    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        out = xp.broadcast_to(
            arr.sum(axis=-1, keepdims=True),
            (*arr.shape[:-1], _.codim),
        )
        return out

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        out = xp.broadcast_to(
            arr.sum(axis=-1, keepdims=True),
            (*arr.shape[:-1], _.dim),
        )
        return out

    def op_gram(_) -> pxt.OpT:
        op = _Sum(M=_.dim, N=_.dim) * _.codim
        return op

    def op_cogram(_) -> pxt.OpT:
        op = _Sum(M=_.codim, N=_.codim) * _.dim
        return op

    if M == 1:
        klass = pxo.LinFunc
    elif M == N:
        klass = pxo.SelfAdjointOp
    else:
        klass = pxo.LinOp
    op = klass(shape=(M, N))
    op._lipschitz = np.sqrt(M * N)
    op.apply = types.MethodType(op_apply, op)
    op.adjoint = types.MethodType(op_adjoint, op)
    op.gram = types.MethodType(op_gram, op)
    op.cogram = types.MethodType(op_cogram, op)
    return op



import collections.abc as cabc
import datetime as dt
import enum
import logging
import operator
import pathlib as plib
import shutil
import sys
import tempfile
import threading
import typing as typ

import numpy as np

import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "SolverMode",
    "Solver",
    "StoppingCriterion",
]


@enum.unique
class SolverMode(enum.Enum):
    """
    Solver execution mode.
    """

    BLOCK = enum.auto()
    MANUAL = enum.auto()
    ASYNC = enum.auto()


class StoppingCriterion:
    """
    State machines (SM) which decide when to stop iterative solvers by examining their mathematical state.

    SM decisions are always accompanied by at least one numerical statistic. These stats may be queried by solvers via
    :py:meth:`~pyxu.abc.StoppingCriterion.info` to provide diagnostic information to users.

    Composite stopping criteria can be implemented via the overloaded (and[``&``], or[``|``]) operators.
    """

    def stop(self, state: cabc.Mapping[str]) -> bool:
        """
        Compute a stop signal based on the current mathematical state.

        Parameters
        ----------
        state: ~collections.abc.Mapping
            Full mathematical state of solver at some iteration, i.e. :py:attr:`~pyxu.abc.Solver._mstate`.

            Values from `state` may be cached inside the instance to form complex stopping conditions.

        Returns
        -------
        s: bool
            True if no further iterations should be performed, False otherwise.
        """
        raise NotImplementedError

    def info(self) -> cabc.Mapping[str, float]:
        """
        Get statistics associated with the last call to :py:meth:`~pyxu.abc.StoppingCriterion.stop`.

        Returns
        -------
        data: ~collections.abc.Mapping
        """
        raise NotImplementedError

    def clear(self):
        """
        Clear SM state (if any).

        This method is useful when a :py:class:`~pyxu.abc.StoppingCriterion` instance must be reused in another call to
        :py:meth:`~pyxu.abc.Solver.fit`.
        """
        pass

    def __or__(self, other: "StoppingCriterion") -> "StoppingCriterion":
        return _StoppingCriteriaComposition(lhs=self, rhs=other, op=operator.or_)

    def __and__(self, other: "StoppingCriterion") -> "StoppingCriterion":
        return _StoppingCriteriaComposition(lhs=self, rhs=other, op=operator.and_)


class _StoppingCriteriaComposition(StoppingCriterion):
    def __init__(
        self,
        lhs: "StoppingCriterion",
        rhs: "StoppingCriterion",
        op: cabc.Callable[[bool, bool], bool],
    ):
        self._lhs = lhs
        self._rhs = rhs
        self._op = op

    def stop(self, state: cabc.Mapping) -> bool:
        return self._op(self._lhs.stop(state), self._rhs.stop(state))

    def info(self) -> cabc.Mapping[str, float]:
        return {**self._lhs.info(), **self._rhs.info()}

    def clear(self):
        self._lhs.clear()
        self._rhs.clear()


class Solver:
    r"""
    Iterative solver for minimization problems of the form :math:`\hat{x} = \arg\min_{x \in \mathbb{R}^{N}}
    \mathcal{F}(x)`, where the form of :math:`\mathcal{F}` is solver-dependent.

    Solver provides a versatile API for solving optimisation problems, with the following features:

    * manual/automatic/background execution of solver iterations via parameters provided to
      :py:meth:`~pyxu.abc.Solver.fit`. (See below.)
    * automatic checkpointing of solver progress, providing a safe restore point in case of faulty numerical code.  Each
      solver instance backs its state and final output to a folder on disk for post-analysis.  In particular
      :py:meth:`~pyxu.abc.Solver.fit` will never crash: detailed exception information will always be available in a
      logfile for post-analysis.
    * arbitrary specification of complex stopping criteria via the :py:class:`~pyxu.abc.StoppingCriterion` class.
    * solve for multiple initial points in parallel.

    To implement a new iterative solver, users need to sub-class :py:class:`~pyxu.abc.Solver` and overwrite the methods
    below:

    * :py:meth:`~pyxu.abc.Solver.__init__`
    * :py:meth:`~pyxu.abc.Solver.m_init`  [i.e. math-init()]
    * :py:meth:`~pyxu.abc.Solver.m_step`  [i.e. math-step()]
    * :py:meth:`~pyxu.abc.Solver.default_stop_crit`  [optional; see method definition for details]
    * :py:meth:`~pyxu.abc.Solver.objective_func`  [optional; see method definition for details]

    Advanced functionalities of :py:class:`~pyxu.abc.Solver` are automatically inherited by sub-classes.


    Examples
    --------
    Here are examples on how to solve minimization problems with this class:

    .. code-block:: python3

       slvr = Solver()

       ### 1. Blocking mode: .fit() does not return until solver has stopped.
       >>> slvr.fit(mode=SolverMode.BLOCK, ...)
       >>> data, hist = slvr.stats()  # final output of solver.

       ### 2. Async mode: solver iterations run in the background.
       >>> slvr.fit(mode=SolverMode.ASYNC, ...)
       >>> print('test')  # you can do something in between.
       >>> slvr.busy()  # or check whether the solver already stopped.
       >>> slvr.stop()  # and preemptively force it to stop.
       >>> data, hist = slvr.stats()  # then query the result after a (potential) force-stop.

       ### 3. Manual mode: fine-grain control of solver data per iteration.
       >>> slvr.fit(mode=SolverMode.MANUAL, ...)
       >>> for data in slvr.steps():
       ...     # Do something with the logged variables after each iteration.
       ...     pass  # solver has stopped after the loop.
       >>> data, hist = slvr.stats()  # final output of solver.
    """

    _mstate: dict[str, typ.Any]  #: Mathematical state.
    _astate: dict[str, typ.Any]  #: Book-keeping (non-math) state.

    def __init__(
        self,
        *,
        folder: pxt.Path = None,
        exist_ok: bool = False,
        stop_rate: pxt.Integer = 1,
        writeback_rate: pxt.Integer = None,
        verbosity: pxt.Integer = None,
        show_progress: bool = True,
        log_var: pxt.VarName = frozenset(),
    ):
        """
        Parameters
        ----------
        folder: Path
            Directory on disk where instance data should be stored.  A location will be automatically chosen if
            unspecified. (Default: OS-dependent tempdir.)
        exist_ok: bool
            If `folder` is specified and `exist_ok` is false (default), :py:class:`FileExistsError` is raised if the
            target directory already exists.
        stop_rate: Integer
            Rate at which solver evaluates stopping criteria.
        writeback_rate: Integer
            Rate at which solver checkpoints are written to disk.  No checkpointing is done if unspecified: only the
            final solver output will be written back to disk.  Must be a multiple of `stop_rate` if specified.
        verbosity: Integer
            Rate at which stopping criteria statistics are logged.  Must be a multiple of `stop_rate`.  Defaults to
            `stop_rate` if unspecified.
        show_progress: bool
            If True (default) and :py:meth:`~pyxu.abc.Solver.fit` is run with mode=BLOCK, then statistics are also
            logged to stdout.
        log_var: VarName
            Variables from the solver's math-state (:py:attr:`~pyxu.abc.Solver._mstate`) to be logged per iteration.
            These are the variables made available when calling :py:meth:`~pyxu.abc.Solver.stats`.

        Notes
        -----
        * Partial device<>CPU synchronization takes place when stopping-criteria are evaluated.  Increasing `stop_rate`
          is advised to reduce the effect of such transfers when applicable.
        * Full device<>CPU synchronization takes place at checkpoint-time.  Increasing `writeback_rate` is advised to
          reduce the effect of such transfers when applicable.
        """
        self._mstate = dict()
        self._astate = dict(
            history=None,  # stopping criteria values per iteration
            idx=0,  # iteration index
            log_rate=None,
            log_var=None,
            logger=None,
            stdout=None,
            stop_crit=None,
            stop_rate=None,
            track_objective=None,
            wb_rate=None,
            workdir=None,
            # Execution-mode related -----------
            mode=None,
            active=None,
            worker=None,
        )

        try:
            if folder is None:
                folder = plib.Path(tempfile.mkdtemp(prefix="pyxu_"))
            elif (folder := plib.Path(folder).expanduser().resolve()).exists() and (not exist_ok):
                raise FileExistsError(f"{folder} already exists.")
            else:
                shutil.rmtree(folder, ignore_errors=True)
                folder.mkdir(parents=True)
            self._astate["workdir"] = folder
        except Exception:
            raise Exception(f"folder: expected path-like, got {type(folder)}.")

        try:
            assert stop_rate >= 1
            self._astate["stop_rate"] = int(stop_rate)
        except Exception:
            raise ValueError(f"stop_rate must be positive, got {stop_rate}.")

        try:
            self._astate["wb_rate"] = writeback_rate
            if writeback_rate is not None:
                assert writeback_rate % self._astate["stop_rate"] == 0
                self._astate["wb_rate"] = int(writeback_rate)
        except Exception:
            raise ValueError(f"writeback_rate must be a multiple of stop_rate({stop_rate}), got {writeback_rate}.")

        try:
            if verbosity is None:
                verbosity = self._astate["stop_rate"]
            assert verbosity % self._astate["stop_rate"] == 0
            self._astate["log_rate"] = int(verbosity)
            self._astate["stdout"] = bool(show_progress)
        except Exception:
            raise ValueError(f"verbosity must be a multiple of stop_rate({stop_rate}), got {verbosity}.")

        try:
            if isinstance(log_var, str):
                log_var = (log_var,)
            self._astate["log_var"] = frozenset(log_var)
        except Exception:
            raise ValueError(f"log_var: expected collection, got {type(log_var)}.")

    def fit(self, **kwargs):
        r"""
        Solve minimization problem(s) defined in :py:meth:`~pyxu.abc.Solver.__init__`, with the provided run-specifc
        parameters.

        Parameters
        ----------
        kwargs
            See class-level docstring for class-specific keyword parameters.
        stop_crit: StoppingCriterion
            Stopping criterion to end solver iterations.  If unspecified, defaults to
            :py:meth:`~pyxu.abc.Solver.default_stop_crit`.
        mode: SolverMode
            Execution mode.
            See :py:class:`~pyxu.abc.Solver` for usage examples.

            Useful method pairs depending on the execution mode:

            * BLOCK: :py:meth:`~pyxu.abc.Solver.fit`
            * ASYNC: :py:meth:`~pyxu.abc.Solver.fit`, :py:meth:`~pyxu.abc.Solver.busy`, :py:meth:`~pyxu.abc.Solver.stop`
            * MANUAL: :py:meth:`~pyxu.abc.Solver.fit`, :py:meth:`~pyxu.abc.Solver.steps`
        track_objective: bool
            Auto-compute objective function every time stopping criterion is evaluated.
        """
        self._fit_init(
            mode=kwargs.pop("mode", SolverMode.BLOCK),
            stop_crit=kwargs.pop("stop_crit", None),
            track_objective=kwargs.pop("track_objective", False),
        )
        self.m_init(**kwargs)
        self._fit_run()

    def m_init(self, **kwargs):
        """
        Set solver's initial mathematical state based on kwargs provided to :py:meth:`~pyxu.abc.Solver.fit`.

        This method must only manipulate :py:attr:`~pyxu.abc.Solver._mstate`.

        After calling this method, the solver must be able to complete its 1st iteration via a call to
        :py:meth:`~pyxu.abc.Solver.m_step`.
        """
        raise NotImplementedError

    def m_step(self):
        """
        Perform one (mathematical) step.

        This method must only manipulate :py:attr:`~pyxu.abc.Solver._mstate`.
        """
        raise NotImplementedError

    def steps(self, n: pxt.Integer = None) -> cabc.Generator:
        """
        Generator of logged variables after each iteration.

        The i-th call to :py:func:`next` on this object returns the logged variables after the i-th solver iteration.

        This method is only usable after calling :py:meth:`~pyxu.abc.Solver.fit` with mode=MANUAL.  See
        :py:class:`~pyxu.abc.Solver` for usage examples.

        There is no guarantee that a checkpoint on disk exists when the generator is exhausted.  (Reason: potential
        exceptions raised during solver's progress.) Users should invoke :py:meth:`~pyxu.abc.Solver.writeback`
        afterwards if needed.

        Parameters
        ----------
        n: Integer
            Maximum number of :py:func:`next` calls allowed before exhausting the generator.  Defaults to infinity if
            unspecified.

            The generator will terminate prematurely if the solver naturally stops before `n` calls to :py:func:`next`
            are made.
        """
        self._check_mode(SolverMode.MANUAL)
        i = 0
        while (n is None) or (i < n):
            if self._step():
                data, _ = self.stats()
                yield data
                i += 1
            else:
                self._astate["mode"] = None  # force steps() to be call-once when exhausted.
                self._cleanup_logger()
                return

    _stats_data_spec = dict[str, typ.Union[pxt.Real, pxt.NDArray, None]]
    _stats_history_spec = typ.Union[np.ndarray, None]

    def stats(self) -> tuple[_stats_data_spec, _stats_history_spec]:
        """
        Query solver state.

        Returns
        -------
        data: ~collections.abc.Mapping
            Value(s) of ``log_var`` (s) after last iteration.
        history: numpy.ndarray, None
            (N_iter,) records of stopping-criteria values per iteration.

        Notes
        -----
        If any of the ``log_var`` (s) and/or ``history`` are not (yet) known at query time, ``None`` is returned.
        """
        history = self._astate["history"]
        if history is not None:
            if len(history) > 0:
                history = np.concatenate(history, dtype=history[0].dtype, axis=0)
            else:
                history = None
        data = {k: self._mstate.get(k) for k in self._astate["log_var"]}
        return data, history

    @property
    def workdir(self) -> pxt.Path:
        """
        Returns
        -------
        wd: Path
            Absolute path to the directory on disk where instance data is stored.
        """
        return self._astate["workdir"]

    @property
    def logfile(self) -> pxt.Path:
        """
        Returns
        -------
        lf: Path
            Absolute path to the log file on disk where stopping criteria statistics are logged.
        """
        return self.workdir / "solver.log"

    @property
    def datafile(self) -> pxt.Path:
        """
        Returns
        -------
        df: Path
            Absolute path to the file on disk where ``log_var`` (s) are stored during checkpointing or after solver has
            stopped.
        """
        return self.workdir / "data.npz"

    def busy(self) -> bool:
        """
        Test if an async-running solver has stopped.

        This method is only usable after calling :py:meth:`~pyxu.abc.Solver.fit` with mode=ASYNC.  See
        :py:class:`~pyxu.abc.Solver` for usage examples.

        Returns
        -------
        b: bool
            True if solver has stopped, False otherwise.
        """
        self._check_mode(SolverMode.ASYNC, SolverMode.BLOCK)
        return self._astate["active"].is_set()

    def solution(self):
        """
        Output the "solution" of the optimization problem.

        This is a helper method intended for novice users.  The return type is sub-class dependent, so don't write an
        API using this: use :py:meth:`~pyxu.abc.Solver.stats` instead.
        """
        raise NotImplementedError

    def stop(self):
        """
        Stop an async-running solver.

        This method is only usable after calling :py:meth:`~pyxu.abc.Solver.fit` with mode=ASYNC.  See
        :py:class:`~pyxu.abc.Solver` for usage examples.

        This method will block until the solver has stopped.

        There is no guarantee that a checkpoint on disk exists once halted.  (Reason: potential exceptions raised during
        solver's progress.) Users should invoke :py:meth:`~pyxu.abc.Solver.writeback` afterwards if needed.

        Users must call this method to terminate an async-solver, even if :py:meth:`~pyxu.abc.Solver.busy` is False.
        """
        self._check_mode(SolverMode.ASYNC, SolverMode.BLOCK)
        self._astate["active"].clear()
        self._astate["worker"].join()
        self._astate.update(
            mode=None,  # forces stop() to be call-once.
            active=None,
            worker=None,
        )
        self._cleanup_logger()

    def _fit_init(
        self,
        mode: SolverMode,
        stop_crit: StoppingCriterion,
        track_objective: bool,
    ):
        def _init_logger():
            log_name = str(self.workdir)
            logger = logging.getLogger(log_name)
            logger.handlers.clear()
            logger.setLevel("DEBUG")

            fmt = logging.Formatter(fmt="{levelname} -- {message}", style="{")
            handler = [logging.FileHandler(self.logfile, mode="w")]
            if (mode is SolverMode.BLOCK) and self._astate["stdout"]:
                handler.append(logging.StreamHandler(sys.stdout))
            for h in handler:
                h.setLevel("DEBUG")
                h.setFormatter(fmt)
                logger.addHandler(h)

            return logger

        self._mstate.clear()

        if stop_crit is None:
            stop_crit = self.default_stop_crit()
        stop_crit.clear()

        if track_objective:
            from pyxu.opt.stop import Memorize

            stop_crit |= Memorize(var="objective_func")

        self._astate.update(  # suitable state for a new call to fit().
            history=[],
            idx=0,
            logger=_init_logger(),
            stop_crit=stop_crit,
            track_objective=track_objective,
            mode=mode,
            active=None,
            worker=None,
        )

    def _fit_run(self):
        self._m_persist()

        mode = self._astate["mode"]
        if mode is SolverMode.MANUAL:
            # User controls execution via steps().
            pass
        else:  # BLOCK / ASYNC
            self._astate.update(
                active=threading.Event(),
                worker=Solver._Worker(self),
            )
            self._astate["active"].set()
            self._astate["worker"].start()
            if mode is SolverMode.BLOCK:
                self._astate["worker"].join()
                self.stop()  # state clean-up
            else:
                # User controls execution via busy() + stop().
                pass

    def writeback(self):
        """
        Checkpoint state to disk.
        """
        data, history = self.stats()
        kwargs = {k: v for (k, v) in dict(history=history, **data).items() if (v is not None)}
        np.savez(self.datafile, **pxu.compute(kwargs))  # savez() requires NumPy arrays as input.
        # [TODO][Feature Request] Allow user to choose writeback format. Useful for large-scale
        # outputs which cannot be stored on one machine.

    def _check_mode(self, *modes: SolverMode):
        m = self._astate["mode"]
        if m in modes:
            pass  # ok
        else:
            if m is None:
                msg = "Illegal method call: invoke Solver.fit() first."
            else:
                msg = " ".join(
                    [
                        "Illegal method call: can only be used if Solver.fit() invoked with",
                        "mode=Any[" + ", ".join(map(lambda _: str(_.name), modes)) + "]",
                    ]
                )
            raise ValueError(msg)

    def _step(self) -> bool:
        ast = self._astate  # shorthand

        must_stop = lambda: ast["idx"] % ast["stop_rate"] == 0
        must_log = lambda: ast["idx"] % ast["log_rate"] == 0
        must_writeback = lambda: (ast["wb_rate"] is not None) and (ast["idx"] % ast["wb_rate"] == 0)

        def _log(msg: str = None):
            if msg is None:  # report stopping-criterion values
                h = ast["history"][-1][0]
                msg = [f"[{dt.datetime.now()}] Iteration {ast['idx']:>_d}"]
                for field, value in zip(h.dtype.names, h):
                    msg.append(f"\t{field}: {value}")
                msg = "\n".join(msg)
            ast["logger"].info(msg)

        def _update_history():
            def _as_struct(data: dict[str, float]) -> np.ndarray:
                ftype = pxrt.getPrecision().value
                spec_data = [(k, ftype) for k in data]

                itype = np.int64
                spec_iter = [("iteration", itype)]

                dtype = np.dtype(spec_iter + spec_data)

                utype = np.uint8
                s = np.concatenate(  # to allow mixed int/float fields:
                    [  # (1) cast to uint, then (2) to compound dtype.
                        np.array([ast["idx"]], dtype=itype).view(utype),
                        np.array(list(data.values()), dtype=ftype).view(utype),
                    ]
                ).view(dtype)
                return s

            h = _as_struct(ast["stop_crit"].info())
            ast["history"].append(h)

        # [Sepand] Important
        # stop_crit.stop(), _update_history(), _log() must always be called in this order.

        try:
            _ms, _ml, _mw = must_stop(), must_log(), must_writeback()

            if _ms and ast["track_objective"]:
                self._mstate["objective_func"] = self.objective_func().reshape(-1)

            if _ms and ast["stop_crit"].stop(self._mstate):
                _update_history()
                _log()
                _log(msg=f"[{dt.datetime.now()}] Stopping Criterion satisfied -> END")
                self.writeback()
                return False
            else:
                if _ms:
                    _update_history()
                if _ml:
                    _log()
                if _mw:
                    self.writeback()
                ast["idx"] += 1
                self.m_step()
                if _ms:
                    self._m_persist()
                return True
        except Exception as e:
            msg = f"[{dt.datetime.now()}] Something went wrong -> EXCEPTION RAISED"
            msg_xtra = f"More information: {self.logfile}."
            print("\n".join([msg, msg_xtra]), file=sys.stderr)
            if ast["wb_rate"] is not None:  # checkpointing enabled
                _, r = divmod(ast["idx"], ast["wb_rate"])
                idx_valid = ast["idx"] - r
                msg_idx = f"Last valid checkpoint done at iteration={idx_valid}."
                msg = "\n".join([msg, msg_idx])
            ast["logger"].exception(msg, exc_info=e)
            return False

    def _m_persist(self):
        # Persist math state to avoid re-eval overhead.
        k, v = zip(*self._mstate.items())
        v = pxu.compute(*v, mode="persist", traverse=False)
        self._mstate.update(zip(k, v))
        # [Sepand] Note:
        # The above evaluation strategy with `traverse=False` chosen since _mstate can hold any type
        # of object.

    def _cleanup_logger(self):
        # Close file-handlers
        log_name = str(self.workdir)
        logger = logging.getLogger(log_name)
        for handler in logger.handlers:
            handler.close()

    def default_stop_crit(self) -> StoppingCriterion:
        """
        Default stopping criterion for solver if unspecified in :py:meth:`~pyxu.abc.Solver.fit` calls.

        Sub-classes are expected to overwrite this method.  If not overridden, then omitting the `stop_crit` parameter
        in :py:meth:`~pyxu.abc.Solver.fit` is forbidden.
        """
        raise NotImplementedError("No default stopping criterion defined.")

    def objective_func(self) -> pxt.NDArray:
        """
        Evaluate objective function given current math state.

        The output array must have shape:

        * (1,) if evaluated at 1 point,
        * (N, 1) if evaluated at N different points.

        Sub-classes are expected to overwrite this method.  If not overridden, then enabling `track_objective` in
        :py:meth:`~pyxu.abc.Solver.fit` is forbidden.
        """
        raise NotImplementedError("No objective function defined.")

    class _Worker(threading.Thread):
        def __init__(self, solver: "Solver"):
            super().__init__()
            self.slvr = solver

        def run(self):
            while self.slvr.busy() and self.slvr._step():
                pass
            self.slvr._astate["active"].clear()



import collections.abc as cabc
import numbers as nb
import pathlib as plib
import typing as typ

import numpy.typing as npt

import pyxu.info.deps as pxd

if typ.TYPE_CHECKING:
    import pyxu.abc.operator as pxo
    import pyxu.abc.solver as pxs

#: Supported dense array types.
NDArray = typ.TypeVar("NDArray", *pxd.supported_array_types())

#: Supported dense array modules.
ArrayModule = typ.TypeVar(
    "ArrayModule",
    *[typ.Literal[_] for _ in pxd.supported_array_modules()],
)

#: Supported sparse array types.
SparseArray = typ.TypeVar("SparseArray", *pxd.supported_sparse_types())

#: Supported sparse array modules.
SparseModule = typ.TypeVar(
    "SparseModule",
    *[typ.Literal[_] for _ in pxd.supported_sparse_modules()],
)

#: Top-level abstract :py:class:`~pyxu.abc.Operator` interface exposed to users.
OpT = typ.TypeVar(
    # This list should be kept in sync with all user-facing operators in `pxo`.
    "OpT",
    "pxo.Operator",
    "pxo.DiffFunc",
    "pxo.ProxDiffFunc",
    "pxo.NormalOp",
    "pxo.ProxFunc",
    "pxo.SquareOp",
    "pxo.QuadraticFunc",
    "pxo.ProjOp",
    "pxo.LinFunc",
    "pxo.PosDefOp",
    "pxo.Map",
    "pxo.Func",
    "pxo.OrthProjOp",
    "pxo.DiffMap",
    "pxo.UnitOp",
    "pxo.SelfAdjointOp",
    "pxo.LinOp",
)

#: :py:class:`~pyxu.abc.Operator` hierarchy class type.
OpC = typ.Type[OpT]  # Operator classes

#: Mathematical properties attached to :py:class:`~pyxu.abc.Operator` objects.
Property = "pxo.Property"

#: Top-level abstract :py:class:`~pyxu.abc.Solver` interface exposed to users.
SolverT = typ.TypeVar("SolverT", bound="pxs.Solver")

#: :py:class:`~pyxu.abc.Solver` hierarchy class type.
SolverC = typ.Type[SolverT]

#: Solver run-modes.
SolverM = typ.TypeVar("SolverM", bound="pxs.SolverMode")

Integer = nb.Integral
Real = nb.Real  #: Alias of :py:class:`numbers.Real`.
DType = npt.DTypeLike  #: :py:attr:`~pyxu.info.ptype.NDArray` dtype specifier.
OpShape = tuple[Integer, Integer]  #: :py:class:`~pyxu.abc.Operator` shape specifier.
NDArrayAxis = typ.Union[Integer, tuple[Integer, ...]]  #: Axis/Axes specifier.
NDArrayShape = typ.Union[Integer, tuple[Integer, ...]]  #: :py:attr:`~pyxu.info.ptype.NDArray` shape specifier.
Path = typ.Union[str, plib.Path]  #: Path-like object.
VarName = typ.Union[str, cabc.Collection[str]]  #: Variable name(s).



import importlib
import time

import dask.array as da
import numpy as np

import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu

importlib.reload(pxo)

if __name__ == "__main__":
    rng = np.random.default_rng(0)
    D, M, N = 3, 500_1, 203_0
    rnd_points = lambda _: rng.normal(scale=rng.uniform(0.25, 0.5, size=(D,)), size=(_, D))
    rnd_offset = lambda: rng.uniform(-1, 1, size=(D,))
    scale = 20
    x = np.concatenate(
        [
            rnd_points(M) + rnd_offset() * scale,
            rnd_points(M) + rnd_offset() * scale,
            rnd_points(M) + rnd_offset() * scale,
            rnd_points(M) + rnd_offset() * scale,
            rnd_points(M) + rnd_offset() * scale,
        ],
        axis=0,
    )
    z = np.concatenate(
        [
            rnd_points(N) + rnd_offset() * scale,
            rnd_points(N) + rnd_offset() * scale,
            rnd_points(N) + rnd_offset() * scale,
            rnd_points(N) + rnd_offset() * scale,
            rnd_points(N) + rnd_offset() * scale,
        ],
        axis=0,
    )
    M, N = map(len, [x, z])

    with pxrt.Precision(pxrt.Width.SINGLE):
        kwargs = dict(
            x=x,
            z=z,
            n_trans=5,
            isign=-1,
            eps=1e-3,
            real=False,
            nthreads=1,
        )
        # Chunked NUFFT ===================================
        A = pxo.NUFFT.type3(**kwargs, chunked=True, parallel=False)
        t = time.time()
        x_chunks, z_chunks = A.auto_chunk(
            max_mem=100,
            max_anisotropy=1,
        )
        t_chunk = time.time() - t
        print("t_chunk:", t_chunk)
        t = time.time()
        A.allocate(x_chunks, z_chunks, direct_eval_threshold=50**2)
        t_alloc = time.time() - t
        print("t_alloc:", t_alloc)
        # Regular NUFFT ===================================
        kwargs.update(nthreads=0, eps=0)  # to avoid memory issues
        B = pxo.NUFFT.type3(**kwargs)

        use_dask = True
        arr = rng.normal(size=(kwargs.get("n_trans") * 3, M))
        if not kwargs.get("real"):
            arr = arr + 1j * rng.normal(size=arr.shape)
        if use_dask:
            arr = da.array(arr, dtype=arr.dtype)

    with pxrt.Precision(pxrt.Width.SINGLE):
        t = time.time()
        A_out_fw = A.apply(pxu.view_as_real(arr))
        t_A = time.time() - t
        print("t_A_fw:", t_A)
        t = time.time()
        B_out_fw = B.apply(pxu.view_as_real(arr))
        t_B = time.time() - t
        print("t_B_fw:", t_B)

        t = time.time()
        A_out_bw = A.adjoint(A_out_fw)
        t_A = time.time() - t
        print("t_A_bw:", t_A)
        t = time.time()
        B_out_bw = B.adjoint(B_out_fw)
        t_B = time.time() - t
        print("t_B_bw:", t_B)

        t = time.time()
        res_fw = (np.linalg.norm(A_out_fw - B_out_fw, axis=-1) / np.linalg.norm(B_out_fw, axis=-1)).max()
        t_fw = time.time() - t
        print("res_fw, t_fw:", float(res_fw), t_fw)
        t = time.time()
        res_bw = (np.linalg.norm(A_out_bw - B_out_bw, axis=-1) / np.linalg.norm(B_out_bw, axis=-1)).max()
        t_bw = time.time() - t
        print("res_bw, t_bw:", float(res_bw), t_bw)



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class TestNullFunc(conftest.LinFuncT):
    @pytest.fixture(params=[1, 10])
    def dim(self, request):
        return request.param

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, dim, request):
        op = pxo.NullFunc(dim=dim)
        return op, *request.param

    @pytest.fixture
    def data_apply(self, dim):
        arr = np.arange(dim)
        out = np.zeros((1,))
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    # Q: Why not apply LinOp's test_interface_jacobian() and rely instead on ProxDiffFunc's?
    # A: Because .jacobian() method is forwarded by NullOp(shape=(1, x)).asop(LinFunc), hence
    #    NullFunc().jacobian() is the original NullOp() object, and not the NullFunc() object.
    #    Modifying .asop() to avoid this behaviour is complex. Moreover it only affects NullFunc,
    #    the only squeezed LinOp which we explicitly need to test for correctness.
    def test_interface_jacobian(self, op, _data_apply):
        self._skip_if_disabled()
        conftest.ProxDiffFuncT.test_interface_jacobian(self, op, _data_apply)



import itertools

import numpy as np
import pytest
import scipy.linalg as splinalg

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


def Q(dim: int) -> pxa.PosDefOp:
    import pyxu_tests.operator.examples.test_posdefop as tc_p

    return tc_p.PSDConvolution(N=dim)


def c(dim: int) -> pxa.LinFunc:
    import pyxu_tests.operator.examples.test_linfunc as tc_l

    return tc_l.ScaledSum(N=dim)


def t() -> float:
    return 1


class TestQuadraticFunc(conftest.QuadraticFuncT):
    from pyxu.operator import IdentityOp, NullFunc

    @pytest.fixture(
        params=itertools.product(
            (
                # dim, Q[init], c[init], t[init], Q[ground-truth], c[ground-truth], t[ground-truth]
                (5, None, None, 0, IdentityOp(5), NullFunc(5), 0),
                (5, Q(5), c(5), t(), Q(5), c(5), t()),
            ),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def Q_param(self, _spec) -> pxt.OpT:
        return _spec[0][4]

    @pytest.fixture
    def c_param(self, _spec) -> pxt.OpT:
        return _spec[0][5]

    @pytest.fixture
    def t_param(self, _spec) -> pxt.Real:
        return _spec[0][6]

    @pytest.fixture
    def spec(self, _spec):
        op = pxa.QuadraticFunc(
            shape=(1, _spec[0][0]),
            Q=_spec[0][1],
            c=_spec[0][2],
            t=_spec[0][3],
        )
        return op, _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec) -> int:
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture
    def data_apply(self, dim, Q_param, c_param, t_param):
        dim = self._sanitize(dim, 3)
        arr = self._random_array((dim,))
        out = 0.5 * (arr @ Q_param.apply(arr)) + c_param.apply(arr) + t_param
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, dim, Q_param, c_param):
        dim = self._sanitize(dim, 3)
        arr = self._random_array((dim,))
        out = Q_param.apply(arr) + c_param.asarray().reshape(-1)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, dim, Q_param, c_param):
        dim = self._sanitize(dim, 3)
        arr, tau = self._random_array((dim,)), 2
        out, *_ = splinalg.lstsq(
            Q_param.asarray() + np.eye(dim) / tau,
            arr / tau - c_param.asarray().reshape(-1),
        )
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )



import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.math as pxm
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "CG",
]


class CG(pxa.Solver):
    r"""
    Conjugate Gradient Method.

    The Conjugate Gradient method solves the minimization problem

    .. math::

       \min_{x\in\mathbb{R}^{N}} \frac{1}{2} \mathbf{x}^{T} \mathbf{A} \mathbf{x} - \mathbf{x}^{T} \mathbf{b},

    where :math:`\mathbf{A}: \mathbb{R}^{N} \to \mathbb{R}^{N}` is a *symmetric* *positive definite* operator, and
    :math:`\mathbf{b} \in \mathbb{R}^{N}`.

    The norm of the `explicit residual <https://www.wikiwand.com/en/Conjugate_gradient_method>`_ :math:`\mathbf
    {r}_{k+1}:=\mathbf{b}-\mathbf{Ax}_{k+1}` is used as the default stopping criterion.  This provides a guaranteed
    level of accuracy both in exact arithmetic and in the presence of round-off errors.  By default, the iterations stop
    when the norm of the explicit residual is smaller than 1e-4.

    Parameters (``__init__()``)
    ---------------------------
    * **A** (:py:class:`~pyxu.abc.PosDefOp`)
      --
      Positive-definite operator :math:`\mathbf{A}: \mathbb{R}^{N} \to \mathbb{R}^{N}`.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.__init__`.

    Parameters (``fit()``)
    ----------------------
    * **b** (:py:attr:`~pyxu.info.ptype.NDArray`)
      --
      (..., N) :math:`\mathbf{b}` terms in the CG cost function.

      All problems are solved in parallel.
    * **x0** (:py:attr:`~pyxu.info.ptype.NDArray`, :py:obj:`None`)
      --
      (..., N) initial point(s).

      Must be broadcastable with `b` if provided.  Defaults to 0.
    * **restart_rate** (:py:attr:`~pyxu.info.ptype.Integer`)
      --
      Number of iterations after which restart is applied.

      By default, a restart is done after 'n' iterations, where 'n' corresponds to the dimension of :math:`\mathbf{A}`.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.fit`.
    """

    def __init__(self, A: pxa.PosDefOp, **kwargs):
        kwargs.update(
            log_var=kwargs.get("log_var", ("x",)),
        )
        super().__init__(**kwargs)

        self._A = A

    @pxrt.enforce_precision(i=("b", "x0"))
    def m_init(
        self,
        b: pxt.NDArray,
        x0: pxt.NDArray = None,
        restart_rate: pxt.Integer = None,
    ):
        mst = self._mstate  # shorthand

        if restart_rate is not None:
            assert restart_rate >= 1
            mst["restart_rate"] = int(restart_rate)
        else:
            mst["restart_rate"] = self._A.dim

        xp = pxu.get_array_module(b)
        if x0 is None:
            mst["b"] = b
            mst["x"] = xp.zeros_like(b)
        elif b.shape == x0.shape:
            # No broadcasting involved
            mst["b"] = b
            mst["x"] = x0.copy()
        else:
            # In-place updates involving b/x won't work if shapes differ -> coerce to largest.
            mst["b"], mst["x"] = xp.broadcast_arrays(b, x0)
            mst["x"] = mst["x"].copy()

        mst["residual"] = mst["b"].copy()
        mst["residual"] -= self._A.apply(mst["x"])
        mst["conjugate_dir"] = mst["residual"].copy()

    def m_step(self):
        mst = self._mstate  # shorthand
        x, r, p = mst["x"], mst["residual"], mst["conjugate_dir"]
        xp = pxu.get_array_module(x)

        Ap = self._A.apply(p)
        rr = pxm.norm(r, ord=2, axis=-1, keepdims=True) ** 2
        alpha = rr / (p * Ap).sum(axis=-1, keepdims=True)
        x += alpha * p

        if pxu.compute(xp.any(rr <= pxrt.Width(rr.dtype).eps())):  # explicit eval
            r[:] = mst["b"]
            r -= self._A.apply(x)
        else:  # implicit eval
            r -= alpha * Ap

        # Because CG can only generate n conjugate vectors in an n-dimensional space, it makes sense
        # to restart CG every n iterations.
        if self._astate["idx"] % mst["restart_rate"] == 0:  # explicit eval
            beta = 0
            r[:] = mst["b"]
            r -= self._A.apply(x)
        else:  # implicit eval
            beta = pxm.norm(r, ord=2, axis=-1, keepdims=True) ** 2 / rr
        p *= beta
        p += r

        # for homogenity with other solver code. Optional in CG due to in-place computations.
        mst["x"], mst["residual"], mst["conjugate_dir"] = x, r, p

    def default_stop_crit(self) -> pxa.StoppingCriterion:
        from pyxu.opt.stop import AbsError

        stop_crit = AbsError(
            eps=1e-4,
            var="residual",
            f=None,
            norm=2,
            satisfy_all=True,
        )
        return stop_crit

    def objective_func(self) -> pxt.NDArray:
        x = self._mstate["x"]
        b = self._mstate["b"]

        f = self._A.apply(x)
        f = pxu.copy_if_unsafe(f)
        f /= 2
        f -= b
        f *= x

        return f.sum(axis=-1, keepdims=True)

    def solution(self) -> pxt.NDArray:
        """
        Returns
        -------
        x: NDArray
            (..., N) solution.
        """
        data, _ = self.stats()
        return data.get("x")



import collections.abc as cabc
import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class SumMixin:
    @pytest.fixture
    def _spec(self, request) -> tuple[pxt.NDArrayShape, pxt.NDArrayAxis]:
        # (arg_shape, axis) raw user input [specified in sub-classes]
        raise NotImplementedError

    @pytest.fixture
    def arg_shape(self, _spec) -> pxt.NDArrayShape:
        arg_shape, _ = _spec  # raw user input
        if not isinstance(arg_shape, cabc.Sequence):
            arg_shape = (arg_shape,)
        return arg_shape  # canonical form

    @pytest.fixture
    def axis(self, arg_shape, _spec) -> pxt.NDArrayAxis:
        _, axis = _spec  # raw user input
        if axis is None:
            axis = tuple(range(len(arg_shape)))
        elif not isinstance(axis, cabc.Sequence):
            axis = (axis,)
        return axis  # canonical form

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(
        self,
        _spec,  # used instead of arg_shape/axis fixtures to get raw user inputs
        request,
    ) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        arg_shape, axis = _spec
        op = pxo.Sum(arg_shape=arg_shape, axis=axis)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, arg_shape, axis) -> pxt.OpShape:
        arg_shape = np.array(arg_shape, dtype=int)
        axis = np.array(axis, dtype=int)

        sum_shape = arg_shape.copy()
        sum_shape[axis] = 1

        dim = arg_shape.prod()
        codim = sum_shape.prod()
        return (codim, dim)

    @pytest.fixture
    def data_apply(self, arg_shape, axis) -> conftest.DataLike:
        rng = np.random.default_rng(51)  # for reproducibility
        arr = rng.normal(size=arg_shape)
        out = arr.sum(axis=axis)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestSumLinOp(SumMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            ((5, 3), 0),
            ((5, 3), 1),
            ((5, 3), -1),
            ((5, 3, 4), (1, 2)),  # multiple axes
            ((5, 3, 4), (0, 2)),  # multiple axes
            ((5, 3, 4), (1, 2)),  # multiple axes
        ]
    )
    def _spec(self, request) -> tuple[pxt.NDArrayShape, pxt.NDArrayAxis]:
        return request.param


class TestSumLinFunc(SumMixin, conftest.LinFuncT):
    @pytest.fixture(
        params=[
            (5, None),
            ((5,), None),
            ((5, 3, 4), None),
            (5, 0),
            ((5,), (0,)),
            ((5, 3, 4), (0, 1, 2)),
        ]
    )
    def _spec(self, request) -> tuple[pxt.NDArrayShape, pxt.NDArrayAxis]:
        return request.param



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class Oblique(pxa.ProjOp):
    # f: \bR^{N} -> \bR^{N}
    #      x     -> (\alpha E_{N,1} + E_{N, N}) x
    def __init__(self, N: int, alpha: float):
        super().__init__(shape=(N, N))
        self._alpha = float(alpha)

    @pxrt.enforce_precision("arr")
    def apply(self, arr):
        out = np.zeros_like(arr)
        out[..., -1] = (self._alpha * arr[..., 0]) + arr[..., -1]
        return out

    @pxrt.enforce_precision("arr")
    def adjoint(self, arr):
        out = np.zeros_like(arr)
        out[..., 0] = self._alpha * arr[..., -1]
        out[..., -1] = arr[..., -1]
        return out


class TestOblique(conftest.ProjOpT):
    @pytest.fixture(
        params=itertools.product(
            ((10, 3.1, Oblique(N=10, alpha=3.1)),),  # dim, alpha, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][2], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def alpha(self, _spec):
        return _spec[0][1]

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture
    def data_apply(self, dim, alpha):
        x = self._random_array((dim,))
        y = np.zeros(dim)
        y[-1] = alpha * x[0] + x[-1]
        return dict(
            in_=dict(arr=x),
            out=y,
        )



import itertools
import warnings

import numpy as np
import pytest

import pyxu.experimental.sampler as pxe_sampler
import pyxu.info.deps as pxd
import pyxu.info.warning as pxw
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.experimental.sampler.conftest as ct


class TestULA(ct.SamplerT):
    @staticmethod
    def _check_no_side_effect(sampler, seed, x0):
        # idea:
        # * eval next(gen_1) once [out_1]
        # * in-place update out_1, ex: scale by constant factor [scale]
        # * eval next(gen_2) with same seed and same x0 [out_2]
        # * assert out_1 == out_2, i.e. input x0 was not modified

        xp = pxu.get_array_module(x0)
        rng_1 = xp.random.default_rng(seed=seed)
        gen_1 = sampler.samples(rng=rng_1, x0=x0)

        scale = 10

        with pxrt.EnforcePrecision(False):
            out_1 = next(gen_1)

            if pxu.copy_if_unsafe(out_1) is not out_1:
                # out_1 is read_only -> safe
                return
            else:
                # out_1 is writeable -> test correctness
                out_gt = out_1.copy()
                out_1 *= scale  # Scale output -> does it change x0 ?
                rng_2 = xp.random.default_rng(seed=seed)
                gen_2 = sampler.samples(rng=rng_2, x0=x0)
                out_2 = next(gen_2)

            try:
                assert xp.allclose(out_gt, out_2)
            except AssertionError:
                # Function is non-transparent, but which backend caused it?
                N = pxd.NDArrayInfo
                ndi = N.from_obj(out_1)
                if ndi == N.CUPY:
                    # warn about CuPy-only non-transparency.
                    msg = "\n".join(
                        [
                            f"{sampler} is not transparent when applied to CuPy starting points.",
                            f"If the same test fails for non-CuPy inputs, then {sampler}'s implementation is at fault -> user fix required.",
                            "If the same test passes for non-CuPy inputs, then this warning can be safely ignored.",
                        ]
                    )
                    warnings.warn(msg, pxw.NonTransparentWarning)
                else:
                    raise

    @pytest.fixture
    def dim(self):
        return 4

    @pytest.fixture
    def x0_np(self, dim):
        stack_dim = 2, 3
        return np.arange(np.prod(stack_dim) * dim).reshape((*stack_dim, dim))

    @pytest.fixture
    def x0(self, x0_np, xp, width):
        return xp.array(x0_np, dtype=width.value)

    @pytest.fixture
    def func(self, dim):
        return 1 / 2 * pxo.SquaredL2Norm(dim=dim)  # Sampling of standard normal distribution

    @pytest.fixture
    def sampler(self, func):
        return pxe_sampler.ULA(f=func)

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, sampler, seed, x0_np, request):
        ndi = request.param[0]
        if (xp := ndi.module()) is None:
            pytest.skip(f"{ndi} unsupported on this machine.")
        width = request.param[1]

        gen = sampler.samples(rng=xp.random.default_rng(seed=seed), x0=xp.array(x0_np, dtype=width.value))
        return gen, *request.param

    @pytest.fixture
    def samples_copy(self, func, sampler, seed, x0_np, xp, width):
        rng_1 = xp.random.default_rng(seed=seed)
        rng_2 = xp.random.default_rng(seed=seed)
        sampler._rng = rng_1  # Reset rng object of sampler
        sampler_copy = pxe_sampler.ULA(f=func)  # Create copy of sampler
        samples_copy = sampler_copy.samples(rng=rng_2, x0=xp.array(x0_np, dtype=width.value))
        return samples_copy

    def test_transparent_samples(self, sampler, seed, x0):
        self._skip_if_disabled()
        self._check_no_side_effect(sampler, seed, x0)


class TestMYULA(TestULA):
    @pytest.fixture
    def prox_func(self, dim):
        return pxo.L1Norm(dim=dim)

    @pytest.fixture
    def sampler(self, func, prox_func):
        return pxe_sampler.MYULA(f=func, g=prox_func)

    @pytest.fixture
    def samples_copy(self, func, prox_func, sampler, seed, x0_np, xp, width):
        rng_1 = xp.random.default_rng(seed=seed)
        rng_2 = xp.random.default_rng(seed=seed)
        sampler._rng = rng_1  # Reset rng object of sampler
        sampler_copy = pxe_sampler.MYULA(f=func, g=prox_func)  # Create copy of sampler
        return sampler_copy.samples(rng=rng_2, x0=xp.array(x0_np, dtype=width.value))



#!/usr/bin/env python3

"""
Setup script.
"""

import configparser
import pathlib as plib

import setuptools


def read_file(path: plib.Path) -> str:
    with open(path, mode="r") as f:
        txt = f.read()
    return txt


def write_file(path: plib.Path, txt: str):
    with open(path, mode="w") as f:
        f.write(txt)


def update_extra_requires(cfg_path: plib.Path):
    # overwrite .CFG file with added targets `complete_gpu`, `complete_no_gpu`.
    cfg = configparser.ConfigParser()
    with open(cfg_path, mode="r") as f:
        cfg.read_file(f)

    # Add aggregate targets
    #   * complete_no_gpu
    #   * complete_gpu
    pkg = {"no_gpu": set(), "gpu": set()}
    pkg_blacklist = {
        "dev",
        "complete_no_gpu",
        "complete_gpu",
    }
    xtra = cfg["options.extras_require"]
    for xtra_name, xtra_values in xtra.items():
        if xtra_name not in pkg_blacklist:
            pkg["gpu"].add(xtra_values)
            if not xtra_name.endswith("_gpu"):
                pkg["no_gpu"].add(xtra_values)
    xtra["complete_no_gpu"] = "".join(pkg["no_gpu"])
    xtra["complete_gpu"] = "".join(pkg["gpu"])

    with open(cfg_path, mode="w") as f:
        cfg.write(f)


cfg_path = plib.Path(__file__).parent / "setup.cfg"
cfg_init = read_file(cfg_path)  # Save setup.cfg original state
update_extra_requires(cfg_path)

setuptools.setup(setup_requires=["pbr"], pbr=True)

write_file(cfg_path, cfg_init)  # Restore setup.cfg to original state



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class ScaledSum(pxa.LinFunc):
    # f: \bR^{M} -> \bR
    #      x     -> cumsum(x).sum()
    def __init__(self, N: int):
        super().__init__(shape=(1, N))
        self.lipschitz = np.sqrt(N * (N + 1) * (2 * N + 1) / 6)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        return arr.cumsum(axis=-1).sum(axis=-1, keepdims=True)

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr):
        xp = pxu.get_array_module(arr)
        out = xp.zeros((*arr.shape[:-1], self.dim), dtype=arr.dtype)
        out[..., :] = xp.arange(self.dim, 0, -1, dtype=arr.dtype)
        out *= arr
        return out


class TestScaledSum(conftest.LinFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, ScaledSum(N=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture
    def data_apply(self, dim):
        x = self._random_array((dim,))
        y = x.cumsum().sum(keepdims=True)
        return dict(
            in_=dict(arr=x),
            out=y,
        )



import collections.abc as cabc
import typing as typ

import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "SubSample",
    "Trim",
]


class SubSample(pxa.LinOp):
    r"""
    Multi-dimensional sub-sampling operator.

    This operator extracts a subset of the input matching the provided subset specifier.
    Its Lipschitz constant is 1.

    Examples
    --------
    .. code-block:: python3

       ### Extract even samples of a 1D signal.
       import pyxu.operator as pxo
       x = np.arange(10)
       S = pxo.SubSample(
             x.shape,
             slice(0, None, 2),
       )
       y = S(x)  # array([0, 2, 4, 6, 8])


    .. code-block:: python3

       ### Extract columns[1, 3, -1] from a 2D matrix
       import pyxu.operator as pxo
       x = np.arange(3 * 40).reshape(3, 40)  # the input
       S = pxo.SubSample(
             x.shape,
             slice(None),  # take all rows
             [1, 3, -1],   # and these columns
       )
       y = S(x.reshape(-1)).reshape(3, 3)  # array([[  1.,   3.,  39.],
                                           #        [ 41.,  43.,  79.],
                                           #        [ 81.,  83., 119.]])

    .. code-block:: python3

       ### Extract all red rows of an (D,H,W) RGB image matching a boolean mask.
       import pyxu.operator as pxo
       x = np.arange(3 * 5 * 4).reshape(3, 5, 4)
       mask = np.r_[True, False, False, True, False]
       S = pxo.SubSample(
             x.shape,
             0,            # red channel
             mask,         # row selector
             slice(None),  # all columns; this field can be omitted.
       )
       y = S(x.reshape(-1)).reshape(1, mask.sum(), 4)  # array([[[ 0.,  1.,  2.,  3.],
                                                       #         [12., 13., 14., 15.]]])
    """
    IndexSpec = typ.Union[
        pxt.Integer,
        cabc.Sequence[pxt.Integer],
        slice,
        pxt.NDArray,  # ints or boolean mask (per dimension)
    ]

    TrimSpec = typ.Union[
        pxt.Integer,
        cabc.Sequence[pxt.Integer],
        cabc.Sequence[tuple[pxt.Integer, pxt.Integer]],
    ]

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        *indices: IndexSpec,
    ):
        """
        Parameters
        ----------
        arg_shape: NDArrayShape
            Shape of the data to be sub-sampled.
        indices: ~pyxu.operator.SubSample.IndexSpec
            Sub-sample specifier per dimension. (See examples.)

            Valid specifiers are:

            * integers
            * lists (or arrays) of indices
            * slices
            * 1D boolean masks

            Unspecified trailing dimensions are not sub-sampled.
        """
        self._arg_shape = pxu.as_canonical_shape(arg_shape)

        assert 1 <= len(indices) <= len(self._arg_shape)
        self._idx = [slice(None)] * len(self._arg_shape)
        for i, idx in enumerate(indices):
            if isinstance(idx, pxt.Integer):
                idx = slice(idx, idx + 1)
            self._idx[i] = idx
        self._idx = tuple(self._idx)

        output = np.broadcast_to(0, self._arg_shape)[self._idx]
        self._sub_shape = np.atleast_1d(output).shape

        super().__init__(shape=(np.prod(self._sub_shape), np.prod(self._arg_shape)))
        self.lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        """
        Sub-sample the data.

        Parameters
        ----------
        arr: NDArray
            (..., arg_shape.prod()) data

        Returns
        -------
        out: NDArray
            (..., sub_shape.prod()) sub-sampled data points.
        """
        sh = arr.shape[:-1]
        arr = arr.reshape(*sh, *self._arg_shape)

        selector = (*[slice(None) for dim in sh], *self._idx)
        out = arr[selector].reshape(*sh, -1)

        out = pxu.read_only(out)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        """
        Up-sample the data.

        Parameters
        ----------
        arr: NDArray
            (..., sub_shape.prod()) data points.

        Returns
        -------
        out: NDArray
            (..., arg_shape.prod()) up-sampled data points. (Zero-filled.)
        """
        sh = arr.shape[:-1]
        arr = arr.reshape(*sh, *self._sub_shape)

        xp = pxu.get_array_module(arr)
        out = xp.zeros((*sh, *self._arg_shape), dtype=arr.dtype)
        selector = (*[slice(None) for dim in sh], *self._idx)
        out[selector] = arr

        out = out.reshape(*sh, -1)
        return out

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = pxa.UnitOp.svdvals(self, **kwargs)
        return D

    def gram(self) -> pxt.OpT:
        @pxrt.enforce_precision(i="arr")
        def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
            _op = _._op
            out = _op.adjoint(_op.apply(arr))
            return out

        op = px_src.from_source(
            cls=pxa.OrthProjOp,
            shape=(self.dim, self.dim),
            embed=dict(_op=self),
            apply=op_apply,
        )
        return op

    def cogram(self) -> pxt.OpT:
        from pyxu.operator.linop.base import IdentityOp

        CG = IdentityOp(dim=self.codim)
        return CG.squeeze()

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = self.adjoint(arr)
        out /= 1 + damp
        return out

    def dagger(self, damp: pxt.Real, **kwargs) -> pxt.OpT:
        op = self.T / (1 + damp)
        return op


def Trim(
    arg_shape: pxt.NDArrayShape,
    trim_width: SubSample.TrimSpec,
) -> pxt.OpT:
    """
    Multi-dimensional trimming operator.

    This operator trims the input array in each dimension according to specified widths.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    trim_width: ~pyxu.operator.SubSample.TrimSpec
        Number of values trimmed from the edges of each axis.
        Multiple forms are accepted:

        * ``int``: trim each dimension's head/tail by `trim_width`.
        * ``tuple[int, ...]``: trim dimension[k]'s head/tail by `trim_width[k]`.
        * ``tuple[tuple[int, int], ...]``: trim dimension[k]'s head/tail by `trim_width[k][0]` / `trim_width[k][1]`
          respectively.

    Returns
    -------
    op: OpT
    """
    arg_shape = tuple(arg_shape)
    N_dim = len(arg_shape)

    # transform `trim_width` to canonical form tuple[tuple[int, int]]
    is_seq = lambda _: isinstance(_, cabc.Sequence)
    if not is_seq(trim_width):  # int-form
        trim_width = ((trim_width, trim_width),) * N_dim
    assert len(trim_width) == N_dim, "arg_shape/trim_width are length-mismatched."
    if not is_seq(trim_width[0]):  # tuple[int, ...] form
        trim_width = tuple((w, w) for w in trim_width)
    else:  # tuple[tuple[int, int], ...] form
        pass

    # translate `trim_width` to `indices` needed for SubSample
    indices = []
    for (w_head, w_tail), dim_size in zip(trim_width, arg_shape):
        s = slice(w_head, dim_size - w_tail)
        indices.append(s)

    op = SubSample(arg_shape, *indices)
    return op



import itertools

import numpy as np
import pytest

import pyxu.info.ptype as pxt
import pyxu.opt.solver as pxsl
import pyxu_tests.opt.solver.conftest as conftest


# We disable AutoInferenceWarning since we do not test convergence speed.
@pytest.mark.filterwarnings("ignore::pyxu.info.warning.AutoInferenceWarning")
class TestPGD(conftest.SolverT):
    @staticmethod
    def spec_data(N: int) -> list[tuple[pxt.SolverC, dict, dict]]:
        klass = [
            pxsl.PGD,
        ]

        funcs = conftest.funcs(N, seed=3)
        stream1 = conftest.generate_funcs(funcs, N_term=1)
        stream2 = conftest.generate_funcs(funcs, N_term=2)
        kwargs_init = [
            *[dict(f=f, g=None) for (f, *_) in stream1],
            *[dict(f=None, g=g) for (g, *_) in stream1],
            *[dict(f=f, g=g) for (f, g) in stream2],
        ]

        kwargs_fit = []
        param_sweep = dict(
            x0=[
                np.full(N, 50),
                np.full((2, N), 50),  # multiple initial points
            ],
            tau=[None],
            acceleration=[True, False],
            d=[50, 75],
        )
        for config in itertools.product(*param_sweep.values()):
            d = dict(zip(param_sweep.keys(), config))
            kwargs_fit.append(d)

        data = itertools.product(klass, kwargs_init, kwargs_fit)
        return list(data)

    @pytest.fixture(params=spec_data(N=7))
    def spec(self, request) -> tuple[pxt.SolverC, dict, dict]:
        klass, kwargs_init, kwargs_fit = request.param
        return klass, kwargs_init, kwargs_fit

    @pytest.fixture
    def cost_function(self, kwargs_init) -> dict[str, pxt.OpT]:
        func = [kwargs_init[k] for k in ("f", "g")]
        func = [f for f in func if f is not None]
        if len(func) == 1:  # f or g
            func = func[0]
        else:  # f and g
            func = func[0] + func[1]
        return dict(x=func)



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class CircularConvolution(pxa.NormalOp):
    # f_{h}: \bR^{N} -> \bR^{N}
    #          x     -> h \circ x, h \in \bR^{N}
    def __init__(self, h: pxt.NDArray):
        N = h.size
        super().__init__(shape=(N, N))
        self.lipschitz = N * (h**2).sum()  # Frobenius norm

        self._h_fw = h.reshape(-1)
        self._h_bw = self._h_fw[[0, *np.arange(1, N)[::-1]]]

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        return self._circ_convolve(self._h_fw, arr)

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr):
        return self._circ_convolve(self._h_bw, arr)

    @staticmethod
    def _circ_convolve(h, x):
        xp = pxu.get_array_module(x)
        fw = lambda _: xp.fft.fft(_, axis=-1)
        bw = lambda _: xp.fft.ifft(_, axis=-1)
        h = xp.array(h, dtype=x.dtype)
        out = bw(fw(h) * fw(x)).real
        return out.astype(x.dtype, copy=False)


class TestCircularConvolution(conftest.NormalOpT):
    filter = conftest.NormalOpT._random_array((10,), seed=2)

    @pytest.fixture(
        params=itertools.product(
            (CircularConvolution(h=filter),),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, request):
        return request.param

    @pytest.fixture
    def data_shape(self):
        return (self.filter.size, self.filter.size)

    @pytest.fixture
    def data_apply(self):
        F = self.filter
        N = F.size
        arr = self._random_array((N,))
        out = np.zeros((N,))
        for n in range(N):
            for k in range(N):
                out[n] += arr[k] * F[n - k % N]
        return dict(
            in_=dict(arr=arr),
            out=out,
        )
